# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import os
from collections import deque
from pathlib import Path
from typing import List, Tuple, Union

import cv2
import numpy as np
import tensorflow as tf
import wget

from fastestimator.backend.zscore import zscore
from fastestimator.util.traceability_util import traceable


def cosine_decay(time: int,
                 cycle_length: int,
                 init_lr: float,
                 min_lr: float = 1e-6,
                 start: int = 1,
                 cycle_multiplier: int = 1):
    """Learning rate cosine decay function (using half of cosine curve).

    This method is useful for scheduling learning rates which oscillate over time:
    ```python
    s = fe.schedule.LRScheduler(model=model, lr_fn=lambda step: cosine_decay(step, cycle_length=3750, init_lr=1e-3))
    fe.Estimator(..., traces=[s])
    ```

    For more information, check out SGDR: https://arxiv.org/pdf/1608.03983.pdf.

    Args:
        time: The current step or epoch during training starting from 1.
        cycle_length: The decay cycle length.
        init_lr: Initial learning rate to decay from.
        min_lr: Minimum learning rate.
        start: The step or epoch to start the decay schedule.
        cycle_multiplier: The factor by which next cycle length will be multiplied.

    Returns:
        lr: learning rate given current step or epoch.
    """
    if time < start:
        lr = init_lr
    else:
        time = time - start + 1
        if cycle_multiplier > 1:
            current_cycle_idx = math.ceil(
                math.log(time * (cycle_multiplier - 1) / cycle_length + 1) / math.log(cycle_multiplier)) - 1
            cumulative = cycle_length * (cycle_multiplier**current_cycle_idx - 1) / (cycle_multiplier - 1)
        elif cycle_multiplier == 1:
            current_cycle_idx = math.ceil(time / cycle_length) - 1
            cumulative = current_cycle_idx * cycle_length
        else:
            raise ValueError("multiplier must be at least 1")
        current_cycle_length = cycle_length * cycle_multiplier**current_cycle_idx
        time_in_cycle = (time - cumulative) / current_cycle_length
        lr = (init_lr - min_lr) / 2 * math.cos(time_in_cycle * math.pi) + (init_lr + min_lr) / 2
    return lr


@traceable()
class ARC:
    """A run-of-the-mill learning rate scheduler.

    Args:
        frequency: invoke frequency in terms of number of epochs.
    """
    def __init__(self, frequency: int = 3) -> None:
        self.frequency = frequency
        assert isinstance(self.frequency, int) and self.frequency > 0
        self.model = self._load_arc_model()
        self.lr_multiplier = {0: 1.618, 1: 1.0, 2: 0.618}
        self.train_loss_one_cycle = []
        self.eval_loss_one_cycle = []
        self.all_train_loss = deque([None] * 3, maxlen=3)
        self.all_eval_loss = deque([None] * 3, maxlen=3)
        self.all_train_lr = deque([None] * 3, maxlen=3)
        self.use_eval_loss = True

    def _load_arc_model(self) -> tf.keras.Model:
        arc_weight_path = self._initialize_arc_weight()
        with tf.device("cpu:0"):  # this ensures ARC model will not occupy gpu memory
            model = self.lstm_stacked()
            model.load_weights(arc_weight_path)
        return model

    @staticmethod
    def _initialize_arc_weight() -> str:
        arc_weight_path = os.path.join(str(Path.home()), "fastestimator_data", "arc_model", "arc.h5")
        if not os.path.exists(arc_weight_path):
            print("FastEstimator - Downloading ARC weights to {}".format(arc_weight_path))
            os.makedirs(os.path.split(arc_weight_path)[0], exist_ok=True)
            wget.download("https://github.com/fastestimator-util/fastestimator-misc/raw/master/resource/arc.h5",
                          out=arc_weight_path)
        return arc_weight_path

    @staticmethod
    def lstm_stacked() -> tf.keras.Model:
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Conv1D(filters=32, kernel_size=5, activation='relu', input_shape=(300, 3)))
        model.add(tf.keras.layers.MaxPooling1D(pool_size=2))
        model.add(tf.keras.layers.Conv1D(filters=64, kernel_size=5, activation='relu'))
        model.add(tf.keras.layers.MaxPooling1D(pool_size=2))
        model.add(tf.keras.layers.LSTM(64, return_sequences=True))
        model.add(tf.keras.layers.LSTM(64))
        model.add(tf.keras.layers.Dense(32, activation='relu'))
        model.add(tf.keras.layers.Dense(3, activation='softmax'))
        return model

    def accumulate_single_train_loss(self, train_loss: float) -> None:
        self.train_loss_one_cycle.append(train_loss)

    def accumulate_single_eval_loss(self, eval_loss: float) -> None:
        self.eval_loss_one_cycle.append(eval_loss)

    def accumulate_all_lrs(self, lr: float) -> None:
        self.all_train_lr.append(lr)

    def gather_multiple_eval_losses(self) -> None:
        self.all_eval_loss.append(self.eval_loss_one_cycle)
        self.eval_loss_one_cycle = []

    def gather_multiple_train_losses(self) -> None:
        self.all_train_loss.append(self.train_loss_one_cycle)
        self.train_loss_one_cycle = []

    def predict_next_multiplier(self) -> float:
        train_loss, missing = self._merge_list(self.all_train_loss)
        train_loss = self._preprocess_train_loss(train_loss, missing)
        val_loss, missing = self._merge_list(self.all_eval_loss)
        val_loss = self._preprocess_val_loss(val_loss, missing)
        train_lr, missing = self._merge_list(self.all_train_lr)
        train_lr = self._preprocess_train_lr(train_lr, missing)
        model_inputs = np.concatenate((train_loss, val_loss, train_lr), axis=1)
        model_inputs = np.expand_dims(model_inputs, axis=0)
        with tf.device("cpu:0"):
            model_pred = self.model(model_inputs, training=False)
        action = np.argmax(model_pred)
        return self.lr_multiplier[action]

    def _preprocess_val_loss(self, val_loss: List[float], missing: int) -> np.ndarray:
        if val_loss:
            target_size = (3 - missing) * 100
            val_loss = np.array(val_loss, dtype="float32")
            val_loss = zscore(val_loss)
            val_loss = cv2.resize(val_loss, (1, target_size), interpolation=cv2.INTER_NEAREST)
            if val_loss.size < 300:
                val_loss = np.pad(val_loss, ((300 - val_loss.size, 0), (0, 0)), mode='constant', constant_values=0.0)
        else:
            val_loss = np.zeros([300, 1], dtype="float32")
        return val_loss

    def _preprocess_train_lr(self, train_lr: List[float], missing: int) -> np.ndarray:
        target_size = (3 - missing) * 100
        train_lr = np.array(train_lr) / train_lr[0]
        train_lr = cv2.resize(train_lr, (1, target_size), interpolation=cv2.INTER_NEAREST)
        if train_lr.size < 300:
            train_lr = np.pad(train_lr, ((300 - train_lr.size, 0), (0, 0)), mode='constant', constant_values=1.0)
        return train_lr

    def _preprocess_train_loss(self, train_loss: List[float], missing: int) -> np.ndarray:
        target_size = (3 - missing) * 100
        train_loss = np.array(train_loss, dtype="float32")
        train_loss = cv2.resize(train_loss, (1, target_size))
        train_loss = zscore(train_loss)
        if train_loss.size < 300:
            train_loss = np.pad(train_loss, ((300 - train_loss.size, 0), (0, 0)), mode='constant', constant_values=0.0)
        return train_loss

    def _merge_list(self, data: List[Union[None, float, List[float]]]) -> Tuple[List[float], int]:
        output = []
        missing = 0
        for item in data:
            if isinstance(item, list):
                output.extend(item)
            elif item:
                output.append(item)
            else:
                missing += 1
        return output, missing



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

from fastestimator.op.tensorop.average import Average
from fastestimator.test.unittest_util import is_equal


class TestAverage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tf_data = [tf.constant([1., 2., 4.]), tf.constant([1., 4., 5.])]
        cls.tf_output = tf.constant([1., 3., 4.5])
        cls.torch_data = [torch.tensor([1., 2., 4.]), torch.tensor([1., 4., 5.])]
        cls.torch_output = torch.tensor([1., 3., 4.5])

    def test_tf_input(self):
        average = Average(inputs='x', outputs='x')
        output = average.forward(data=self.tf_data, state={})
        self.assertTrue(is_equal(output, self.tf_output))

    def test_torch_input(self):
        average = Average(inputs='x', outputs='x')
        output = average.forward(data=self.torch_data, state={})
        self.assertTrue(is_equal(output, self.torch_output))



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import contextlib
import io
import os
import shutil
import tempfile
import unittest
from contextlib import closing

from fastestimator.summary.history import HistoryReader, HistoryRecorder, connect, update_settings
from fastestimator.test.unittest_util import sample_system_object


class TestHistoryRecorder(unittest.TestCase):
    def setUp(self):
        self.db_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.db_dir, 'tmp.db')

    def tearDown(self):
        shutil.rmtree(self.db_dir)

    def test_happy_path(self):
        system = sample_system_object()
        recorder = HistoryRecorder(system=system, est_path="test.py", db_path=self.db_path)
        with recorder:
            print("Test Log Capture")
            print("Line 2")
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM history WHERE pk = (?)", [system.exp_id])
            results = cursor.fetchall()
        with self.subTest("History Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("File name captured"):
            self.assertEqual(results['file'], 'test.py')
        with self.subTest("Status updated"):
            self.assertEqual(results['status'], 'Completed')
        # Logs
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM logs WHERE fk = (?)", [system.exp_id])
            results = cursor.fetchall()
        with self.subTest("Log Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("Complete log captured"):
            self.assertListEqual(results['log'].splitlines(), ["Test Log Capture", "Line 2"])
        db.close()

    def test_error_raised(self):
        system = sample_system_object()
        recorder = HistoryRecorder(system=system, est_path="test.py", db_path=self.db_path)
        try:
            with recorder:
                print("Test Log Capture")
                print("Line 2")
                raise RuntimeError("Training Died")
        except RuntimeError:
            pass
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM history WHERE pk = (?)", [system.exp_id])
            results = cursor.fetchall()
        with self.subTest("History Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("File name captured"):
            self.assertEqual(results['file'], 'test.py')
        with self.subTest("Status updated"):
            self.assertEqual(results['status'], 'Failed')
        # Logs
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM logs WHERE fk = (?)", [system.exp_id])
            results = cursor.fetchall()
        with self.subTest("Log Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("Complete log captured"):
            self.assertEqual(results['log'], "Test Log Capture\nLine 2\n")
        # Error
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM errors WHERE fk = (?)", [system.exp_id])
            results = cursor.fetchall()
        with self.subTest("Error Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("Error info captured"):
            self.assertEqual(results['exc_type'], "RuntimeError")
        db.close()

    def test_restore_training(self):
        system1 = sample_system_object()
        recorder1 = HistoryRecorder(system=system1, est_path="test.py", db_path=self.db_path)
        try:
            with recorder1:
                print("Test Log Capture")
                print("Line 2")
                raise RuntimeError("Training Died")
        except RuntimeError:
            pass
        system2 = sample_system_object()
        recorder2 = HistoryRecorder(system=system2, est_path="test.py", db_path=self.db_path)
        with recorder2:
            # Fake a restore wizard
            system2.__dict__.update(system1.__dict__)
            print("Line 3")
            print("Line 4")
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM history")
            results = cursor.fetchall()
        with self.subTest("History Captured and Consolidated"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("File name captured"):
            self.assertEqual(results['file'], 'test.py')
        with self.subTest("Status updated"):
            self.assertEqual(results['status'], 'Completed')
        with self.subTest("Correct PK"):
            self.assertEqual(results['pk'], system1.exp_id)
        with self.subTest("Restarts Incremented"):
            self.assertEqual(results['n_restarts'], 1)
        # Logs
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM logs WHERE fk = (?)", [system1.exp_id])
            results = cursor.fetchall()
        with self.subTest("Log Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("Complete log captured"):
            self.assertListEqual(results['log'].splitlines(), ["Test Log Capture", "Line 2", "Line 3", "Line 4"])
        db.close()

    def test_restore_training_old_missing(self):
        system1 = sample_system_object()
        recorder1 = HistoryRecorder(system=system1, est_path="test.py", db_path=self.db_path)
        try:
            with recorder1:
                print("Test Log Capture")
                print("Line 2")
                raise RuntimeError("Training Died")
        except RuntimeError:
            pass
        db = connect(self.db_path)
        db.execute("DELETE FROM history WHERE pk = (?)", [system1.exp_id])
        db.commit()
        system2 = sample_system_object()
        recorder2 = HistoryRecorder(system=system2, est_path="test.py", db_path=self.db_path)
        with recorder2:
            # Fake a restore wizard
            system2.__dict__.update(system1.__dict__)
            print("Line 3")
            print("Line 4")
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM history")
            results = cursor.fetchall()
        with self.subTest("History Captured and Consolidated"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("File name captured"):
            self.assertEqual(results['file'], 'test.py')
        with self.subTest("Status updated"):
            self.assertEqual(results['status'], 'Completed')
        with self.subTest("Correct PK"):
            self.assertEqual(results['pk'], system1.exp_id)
        with self.subTest("Restarts Incremented"):
            self.assertEqual(results['n_restarts'], 1)
        # Logs
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM logs WHERE fk = (?)", [system1.exp_id])
            results = cursor.fetchall()
        with self.subTest("Log Captured"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("Complete log captured"):
            self.assertListEqual(results['log'].splitlines(), ["Line 3", "Line 4"])
        db.close()

    def test_auto_delete(self):
        update_settings(n_keep=5, db_path=self.db_path)

        for i in range(7):
            system = sample_system_object()
            recorder = HistoryRecorder(system=system, est_path=f"{i}", db_path=self.db_path)
            with recorder:
                print(f"Run {i}")

        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM history")
            results = cursor.fetchall()

        with self.subTest("Ensure correct number retained"):
            self.assertEqual(len(results), 5)
        with self.subTest("Ensure correct entries retained"):
            actual_names = {result['file'] for result in results}
            expected_names = {f"{i}" for i in range(2, 7)}
            self.assertSetEqual(actual_names, expected_names)
        db.close()


class TestHistoryReader(unittest.TestCase):
    def setUp(self):
        self.db_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.db_dir, 'tmp.db')

    def tearDown(self):
        shutil.rmtree(self.db_dir)

    def test_read_basic_with_data(self):
        for i in range(10):
            system = sample_system_object()
            recorder = HistoryRecorder(system=system, est_path=f"{i}", db_path=self.db_path)
            with recorder:
                print("Test Log Capture")
                print("Line 2")
        self.test_read_basic()

    def test_read_basic(self):
        # Ensure none of the options generate invalid queries
        with HistoryReader(db_path=self.db_path) as reader:
            with self.subTest("No Options"):
                reader.read_basic()
            with self.subTest("Limit 20"):
                reader.read_basic(limit=20)
            with self.subTest("CSV"):
                reader.read_basic(as_csv=True)
            with self.subTest("Args"):
                reader.read_basic(include_args=True)
            with self.subTest("Pks"):
                reader.read_basic(include_pk=True)
            # Query Altering
            with self.subTest("Errs"):
                reader.read_basic(errors=True)
            with self.subTest("Features"):
                reader.read_basic(include_features=True)
            with self.subTest("Traces"):
                reader.read_basic(include_traces=True)
            with self.subTest("Datasets"):
                reader.read_basic(include_datasets=True)
            with self.subTest("Pipeline"):
                reader.read_basic(include_pipeline=True)
            with self.subTest("Network"):
                reader.read_basic(include_network=True)
            # Combos
            with self.subTest("Everything"):
                reader.read_basic(errors=True,
                                  include_features=True,
                                  include_traces=True,
                                  include_datasets=True,
                                  include_pipeline=True,
                                  include_network=True)
            with self.subTest("All -Errors"):
                with self.subTest("Everything"):
                    reader.read_basic(errors=False,
                                      include_features=True,
                                      include_traces=True,
                                      include_datasets=True,
                                      include_pipeline=True,
                                      include_network=True)

    def test_read_sql(self):
        with io.StringIO() as buf:
            with contextlib.redirect_stdout(buf):
                with HistoryReader(db_path=self.db_path) as reader:
                    reader.read_sql(query='SELECT COUNT(*) AS count FROM settings', as_csv=True)
            output = buf.getvalue()
            self.assertRegex(output, '.*count\r\n1\r\n.*')



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend import check_nan
from fastestimator.test.unittest_util import is_equal


class TestCheckNaN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data_np_nan = np.array([[[1.0, 2.0], [3.0, np.NaN]], [[5.0, 6.0], [7.0, 8.0]]])
        cls.data_np_inf = np.array([[[1.0, 2.0], [3.0, np.Inf]], [[5.0, 6.0], [7.0, 8.0]]])
        cls.data_tf_nan = tf.constant([[[1.0, 2.0], [3.0, 4.0]], [[np.NaN, 6.0], [7.0, 8.0]]])
        cls.data_tf_inf = tf.constant([[[1.0, 2.0], [3.0, 4.0]], [[np.Inf, 6.0], [7.0, 8.0]]])
        cls.data_torch_nan = torch.tensor([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [np.NaN, 8.0]]])
        cls.data_torch_inf = torch.tensor([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [np.Inf, 8.0]]])
        cls.op_np = np.bool_(True)
        cls.op_tf = tf.constant(True)
        cls.op_torch = torch.tensor(True)

    def test_check_nan_np_value(self):
        with self.subTest('Detect NaN values'):
            self.assertTrue(is_equal(check_nan(self.data_np_nan), self.op_np))
        with self.subTest('Detect Inf values'):
            self.assertTrue(is_equal(check_nan(self.data_np_inf), self.op_np))

    def test_check_nan_tf_value(self):
        with self.subTest('Detect NaN values'):
            self.assertTrue(is_equal(check_nan(self.data_tf_nan), self.op_tf))
        with self.subTest('Detect Inf values'):
            self.assertTrue(is_equal(check_nan(self.data_tf_inf), self.op_tf))

    def test_check_nan_torch_value(self):
        with self.subTest('Detect NaN values'):
            self.assertTrue(is_equal(check_nan(self.data_torch_nan), self.op_torch))
        with self.subTest('Detect Inf values'):
            self.assertTrue(is_equal(check_nan(self.data_torch_inf), self.op_torch))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import unittest

from fastestimator.schedule.lr_shedule import cosine_decay


class TestLRSchedule(unittest.TestCase):
    def test_cosine_decay(self):
        learning_rate = cosine_decay(time=5, cycle_length=10, init_lr=0.01, min_lr=0.0)
        self.assertEqual(learning_rate, 0.005)

    def test_cosine_decay_cycle(self):
        learning_rate = cosine_decay(time=1001, cycle_length=1000, init_lr=0.01, min_lr=0.0)
        self.assertTrue(math.isclose(learning_rate, 0.01, rel_tol=1e-3))

    def test_cosine_decay_multiplier_mid_cycle(self):
        learning_rate = cosine_decay(5, cycle_length=10, init_lr=0.01, min_lr=0.0, cycle_multiplier=2)
        self.assertEqual(learning_rate, 0.005)

    def test_cosine_decay_multiplier_end_cycle(self):
        learning_rate = cosine_decay(10, cycle_length=10, init_lr=0.01, min_lr=0.0, cycle_multiplier=2)
        self.assertEqual(learning_rate, 0.0)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Tuple, Union

from albumentations.augmentations.transforms import ColorJitter as ColorJitterAlb

from fastestimator.op.numpyop.univariate.univariate import ImageOnlyAlbumentation


class ColorJitter(ImageOnlyAlbumentation):
    """Randomly changes the brightness, contrast, saturation and hue of an image.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        brightness: How much to jitter brightness. brightness_factor is chosen uniformly from
            [max(0, 1 - brightness), 1 + brightness] or the given [min, max]. Should be non negative numbers.
        contrast: How much to jitter contrast. contrast_factor is chosen uniformly from
            [max(0, 1 - contrast), 1 + contrast] or the given [min, max]. Should be non negative numbers.
        saturation: How much to jitter saturation. saturation_factor is chosen uniformly from
            [max(0, 1 - saturation), 1 + saturation] or the given [min, max]. Should be non negative numbers.
        hue: How much to jitter hue. hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
            Should have 0 <= hue <= 0.5 or -0.5 <= min <= max <= 0.5.

    Image types:
        uint8, float32
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 brightness: Union[float, Tuple[float]] = 0.2,
                 contrast: Union[float, Tuple[float]] = 0.2,
                 saturation: Union[float, Tuple[float]] = 0.2,
                 hue: Union[float, Tuple[float]] = 0.2):
        super().__init__(
            ColorJitterAlb(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue, always_apply=True),
            inputs=inputs,
            outputs=outputs,
            mode=mode,
            ds_id=ds_id)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import zipfile
from pathlib import Path
from typing import Optional, Tuple

import wget

from fastestimator.dataset.batch_dataset import BatchDataset
from fastestimator.dataset.dir_dataset import DirDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def load_data(batch_size: int, root_dir: Optional[str] = None) -> Tuple[BatchDataset, BatchDataset]:
    """Load and return the horse2zebra dataset.

    Sourced from https://people.eecs.berkeley.edu/~taesung_park/CycleGAN/datasets/horse2zebra.zip. This method will
        download the data to local storage if the data has not been previously downloaded.

    Args:
        batch_size: The desired batch size.
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.

    Returns:
        (train_data, eval_data)
    """
    home = str(Path.home())

    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'horse2zebra')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'horse2zebra')
    os.makedirs(root_dir, exist_ok=True)

    data_compressed_path = os.path.join(root_dir, 'horse2zebra.zip')
    data_folder_path = os.path.join(root_dir, 'images')

    if not os.path.exists(data_folder_path):
        # download
        if not os.path.exists(data_compressed_path):
            print("Downloading data to {}".format(root_dir))
            wget.download('https://people.eecs.berkeley.edu/~taesung_park/CycleGAN/datasets/horse2zebra.zip',
                          root_dir,
                          bar=bar_custom)

        # extract
        print("\nExtracting files ...")
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            zip_file.extractall(root_dir)
        os.rename(os.path.join(root_dir, 'horse2zebra'), data_folder_path)

    test_a = DirDataset(root_dir=os.path.join(data_folder_path, 'testA'),
                        data_key="A",
                        file_extension='.jpg',
                        recursive_search=False)
    test_b = DirDataset(root_dir=os.path.join(data_folder_path, 'testB'),
                        data_key="B",
                        file_extension='.jpg',
                        recursive_search=False)
    train_a = DirDataset(root_dir=os.path.join(data_folder_path, 'trainA'),
                         data_key="A",
                         file_extension='.jpg',
                         recursive_search=False)
    train_b = DirDataset(root_dir=os.path.join(data_folder_path, 'trainB'),
                         data_key="B",
                         file_extension='.jpg',
                         recursive_search=False)
    outputs = (BatchDataset(datasets=[train_a, train_b], num_samples=[batch_size, batch_size]),
               BatchDataset(datasets=[test_a, test_b], num_samples=[batch_size, batch_size]))
    return outputs



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.test.unittest_util import is_equal, sample_system_object
from fastestimator.trace.metric import CalibrationError
from fastestimator.util import Data


class TestCalibrationError(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.calibration_error = CalibrationError(true_key='y', pred_key='y_pred')
        cls.calibration_error.system = sample_system_object()

    def test_on_epoch_begin(self):
        self.calibration_error.on_epoch_begin(data=Data())
        with self.subTest('Check initial value of y_true'):
            self.assertEqual(self.calibration_error.y_true, [])
        with self.subTest('Check initial value of y_pred'):
            self.assertEqual(self.calibration_error.y_pred, [])

    def test_on_batch_end(self):
        self.calibration_error.y_true = []
        self.calibration_error.y_pred = []
        batch1 = {'y': np.array([0, 0, 1, 1]), 'y_pred': np.array([[1.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.0, 1.0]])}
        self.calibration_error.on_batch_end(data=Data(batch1))
        with self.subTest('Check true values'):
            self.assertTrue(is_equal(self.calibration_error.y_true, list(batch1['y'])))
        with self.subTest('Check pred values'):
            self.assertTrue(is_equal(self.calibration_error.y_pred, list(batch1['y_pred'])))
        batch2 = {'y': np.array([1, 1, 0, 0]), 'y_pred': np.array([[0.0, 1.0], [0.0, 1.0], [1.0, 0.0], [1.0, 0.0]])}
        self.calibration_error.on_batch_end(data=Data(batch2))
        with self.subTest('Check true values (2 batches)'):
            self.assertTrue(is_equal(self.calibration_error.y_true, list(batch1['y']) + list(batch2['y'])))
        with self.subTest('Check pred values (2 batches)'):
            self.assertTrue(is_equal(self.calibration_error.y_pred, list(batch1['y_pred']) + list(batch2['y_pred'])))

    def test_on_epoch_end(self):
        self.calibration_error.y_true = [0] * 50 + [1] * 50
        self.calibration_error.y_pred = list(np.array([1.0, 0.0] * 50 + [0.0, 1.0] * 50).reshape(100, 2))
        data = Data()
        self.calibration_error.on_epoch_end(data=data)
        with self.subTest('Check if calibration error exists'):
            self.assertIn('calibration_error', data)
        with self.subTest('Check the value of calibration error'):
            self.assertEqual(0.0, data['calibration_error'])

    def test_perfect_calibration(self):
        self.calibration_error.y_true = [0] * 50 + [1] * 50
        self.calibration_error.y_pred = list(
            np.array([1.0, 0.0] * 25 + [0.5, 0.5] * 50 + [0.0, 1.0] * 25).reshape(100, 2))
        data = Data()
        self.calibration_error.on_epoch_end(data=data)
        self.assertEqual(0.0, data['calibration_error'])

    def test_imperfect_calibration(self):
        self.calibration_error.y_true = [0] * 50 + [1] * 50
        self.calibration_error.y_pred = list(np.array([1.0, 0.0] * 50 + [0.5, 0.5] * 50).reshape(100, 2))
        data = Data()
        self.calibration_error.on_epoch_end(data=data)
        self.assertEqual(0.3536, data['calibration_error'])



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tarfile
from pathlib import Path
from typing import Optional

import wget

from fastestimator.dataset.dir_dataset import DirDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def _download_data(link: str, data_path: str, idx: int, total_idx: int) -> None:
    """A helper function to run wget.

    Args:
        link: The download link.
        data_path: Where to save the downloaded data.
        idx: The current download index.
        total_idx: How many files total will be downloaded.
    """
    if not os.path.exists(data_path):
        print("Downloading data to {}, file: {} / {}".format(data_path, idx + 1, total_idx))
        wget.download(link, data_path, bar=bar_custom)


def load_data(root_dir: Optional[str] = None) -> DirDataset:
    """Load and return the NIH Chest X-ray dataset.

    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.

    Returns:
        train_data
    """
    if root_dir is None:
        root_dir = os.path.join(str(Path.home()), 'fastestimator_data', 'NIH_Chestxray')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'NIH_Chestxray')
    os.makedirs(root_dir, exist_ok=True)

    image_extracted_path = os.path.join(root_dir, 'images')

    if not os.path.exists(image_extracted_path):
        # download data
        links = [
            'https://nihcc.box.com/shared/static/vfk49d74nhbxq3nqjg0900w5nvkorp5c.gz',
            'https://nihcc.box.com/shared/static/i28rlmbvmfjbl8p2n3ril0pptcmcu9d1.gz',
            'https://nihcc.box.com/shared/static/f1t00wrtdk94satdfb9olcolqx20z2jp.gz',
            'https://nihcc.box.com/shared/static/0aowwzs5lhjrceb3qp67ahp0rd1l1etg.gz',
            'https://nihcc.box.com/shared/static/v5e3goj22zr6h8tzualxfsqlqaygfbsn.gz',
            'https://nihcc.box.com/shared/static/asi7ikud9jwnkrnkj99jnpfkjdes7l6l.gz',
            'https://nihcc.box.com/shared/static/jn1b4mw4n6lnh74ovmcjb8y48h8xj07n.gz',
            'https://nihcc.box.com/shared/static/tvpxmn7qyrgl0w8wfh9kqfjskv6nmm1j.gz',
            'https://nihcc.box.com/shared/static/upyy3ml7qdumlgk2rfcvlb9k6gvqq2pj.gz',
            'https://nihcc.box.com/shared/static/l6nilvfa9cg3s28tqv1qc1olm3gnz54p.gz',
            'https://nihcc.box.com/shared/static/hhq8fkdgvcari67vfhs7ppg2w6ni4jze.gz',
            'https://nihcc.box.com/shared/static/ioqwiy20ihqwyr8pf4c24eazhh281pbu.gz'
        ]
        data_paths = [os.path.join(root_dir, "images_{}.tar.gz".format(x)) for x in range(len(links))]
        for idx, (link, data_path) in enumerate(zip(links, data_paths)):
            _download_data(link, data_path, idx, len(links))

        # extract data
        for idx, data_path in enumerate(data_paths):
            print("Extracting {}, file {} / {}".format(data_path, idx + 1, len(links)))
            with tarfile.open(data_path) as img_tar:
                img_tar.extractall(root_dir)

    return DirDataset(image_extracted_path, file_extension='.png', recursive_search=False)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import random
from collections import defaultdict
from copy import deepcopy
from functools import lru_cache
from typing import Any, Dict, Hashable, Iterable, List, Optional, Sequence, Tuple, Union

import jsonpickle
import numpy as np
from torch.utils.data import Dataset

from fastestimator.util.traceability_util import FeSplitSummary, traceable
from fastestimator.util.util import FEID, get_shape, get_type


class KeySummary:
    """A summary of the dataset attributes corresponding to a particular key.

    This class is intentionally not @traceable.

    Args:
        num_unique_values: The number of unique values corresponding to a particular key (if known).
        shape: The shape of the vectors corresponding to the key. None is used in a list to indicate that a dimension is
            ragged.
        dtype: The data type of instances corresponding to the given key.
    """
    num_unique_values: Optional[int]
    shape: List[Optional[int]]
    dtype: str

    def __init__(self, dtype: str, num_unique_values: Optional[int] = None, shape: List[Optional[int]] = ()) -> None:
        self.num_unique_values = num_unique_values
        self.shape = shape
        self.dtype = dtype

    def __repr__(self):
        return "<KeySummary {}>".format(self.__getstate__())

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


class DatasetSummary:
    """This class contains information summarizing a dataset object.

    This class is intentionally not @traceable.

    Args:
        num_instances: The number of data instances within the dataset (influences the size of an epoch).
        num_classes: How many different classes are present.
        keys: What keys does the dataset provide, along with summary information about each key.
        class_key: Which key corresponds to class information (if known).
        class_key_mapping: A mapping of the original class string values to the values which are output to the pipeline.
    """
    num_instances: int
    num_classes: Optional[int]
    class_key: Optional[str]
    class_key_mapping: Optional[Dict[str, Any]]
    keys: Dict[str, KeySummary]

    def __init__(self,
                 num_instances: int,
                 keys: Dict[str, KeySummary],
                 num_classes: Optional[int] = None,
                 class_key_mapping: Optional[Dict[str, Any]] = None,
                 class_key: Optional[str] = None):
        self.num_instances = num_instances
        self.class_key = class_key
        self.num_classes = num_classes
        self.class_key_mapping = class_key_mapping
        self.keys = keys

    def __repr__(self):
        return "<DatasetSummary {}>".format(self.__getstate__())

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def __str__(self):
        return jsonpickle.dumps(self, unpicklable=False)


@traceable()
class FEDataset(Dataset):
    def __len__(self) -> int:
        """Defines how many datapoints the dataset contains.

        This is used for computing the number of datapoints available per epoch.

        Returns:
            The number of datapoints within the dataset.
        """
        raise NotImplementedError

    def __getitem__(self, index: int) -> Dict[str, Any]:
        """Fetch a data instance at a specified index.

        Args:
            index: Which datapoint to retrieve.

        Returns:
            The data dictionary from the specified index.
        """
        raise NotImplementedError

    @classmethod
    def fix_split_traceabilty(cls,
                              parent: 'FEDataset',
                              children: List['FEDataset'],
                              fractions: Tuple[Union[float, int, Iterable[int]], ...],
                              seed: Optional[int],
                              stratify: Optional[str]) -> None:
        """A method to fix traceability information after invoking the dataset .split() method.

        Note that the default implementation of the .split() function invokes this already, so this only needs to be
        invoked if you override the .split() method when defining a subclass (ex. BatchDataset).

        Args:
            parent: The parent dataset on which .split() was invoked.
            children: The datasets generated by performing the split.
            fractions: The fraction arguments used to generate the children (should be one-to-one with the children).
            seed: The random seed used to generate the split.
            stratify: The stratify key used to generate the split.
        """
        if hasattr(parent, '_fe_traceability_summary'):
            parent_id = FEID(id(parent))
            fractions = [
                f"range({frac.start}, {frac.stop}, {frac.step})" if isinstance(frac, range) else f"{frac}"
                for frac in fractions
            ]
            for child, frac in zip(children, fractions):
                # noinspection PyProtectedMember
                tables = deepcopy(child._fe_traceability_summary)
                # Update the ID if necessary
                child_id = FEID(id(child))
                if child_id not in tables:
                    # The child was created without invoking its __init__ method, so its internal summary will have the
                    # wrong id
                    table = tables.pop(parent_id)
                    table.fe_id = child_id
                    tables[child_id] = table
                else:
                    table = tables[child_id]
                split_summary = table.fields.get('split', FeSplitSummary())
                split_summary.add_split(parent=parent_id, fraction=frac, seed=seed, stratify=stratify)
                table.fields['split'] = split_summary
                child._fe_traceability_summary = tables
            # noinspection PyUnresolvedReferences
            table = parent._fe_traceability_summary.get(parent_id)
            split_summary = table.fields.get('split', FeSplitSummary())
            split_summary.add_split(parent='self',
                                    fraction=", ".join([f"-{frac}" for frac in fractions]),
                                    seed=seed,
                                    stratify=stratify)
            table.fields['split'] = split_summary
            # Put the new parent summary into the child table to ensure it will always exist in the final set of tables
            for child in children:
                child._fe_traceability_summary[parent_id] = deepcopy(table)

    def split(self,
              *fractions: Union[float, int, Iterable[int]],
              seed: Optional[int] = None,
              stratify: Optional[str] = None) -> Union['FEDataset', List['FEDataset']]:
        """Split this dataset into multiple smaller datasets.

        This function enables several types of splitting:
        1. Splitting by fractions.
            ```python
            ds = fe.dataset.FEDataset(...)  # len(ds) == 1000
            ds2 = ds.split(0.1)  # len(ds) == 900, len(ds2) == 100
            ds3, ds4 = ds.split(0.1, 0.2)  # len(ds) == 630, len(ds3) == 90, len(ds4) == 180
            ```
        2. Splitting by counts.
            ```python
            ds = fe.dataset.FEDataset(...)  # len(ds) == 1000
            ds2 = ds.split(100)  # len(ds) == 900, len(ds2) == 100
            ds3, ds4 = ds.split(90, 180)  # len(ds) == 630, len(ds3) == 90, len(ds4) == 180
            ```
        3. Splitting by indices.
            ```python
            ds = fe.dataset.FEDataset(...)  # len(ds) == 1000
            ds2 = ds.split([87,2,3,100,121,158])  # len(ds) == 994, len(ds2) == 6
            ds3 = ds.split(range(100))  # len(ds) == 894, len(ds3) == 100
            ```

        Args:
            *fractions: Floating point values will be interpreted as percentages, integers as an absolute number of
                datapoints, and an iterable of integers as the exact indices of the data that should be removed in order
                to create the new dataset.
            seed: The random seed to use when splitting the dataset. Useful if you want consistent splits across
                multiple experiments. This isn't necessary if you are splitting by data index.
            stratify: A class key within the dataset with which to stratify the split (to approximately maintain class
                balance ratios before and after a split). Incompatible with data index splitting.

        Returns:
            One or more new datasets which are created by removing elements from the current dataset. The number of
            datasets returned will be equal to the number of `fractions` provided. If only a single value is provided
            then the return will be a single dataset rather than a list of datasets.

        Raises:
            AssertionError: If input arguments are unacceptable.
        """
        assert len(fractions) > 0, "split requires at least one fraction argument"
        original_size = self._split_length()
        method = None
        frac_sum = 0
        int_sum = 0
        n_samples = []
        for frac in fractions:
            if isinstance(frac, float):
                frac_sum += frac
                frac = math.ceil(original_size * frac)
                int_sum += frac
                n_samples.append(frac)
                if method is None:
                    method = 'number'
                assert method == 'number', "Split supports either numeric splits or lists of indices but not both"
            elif isinstance(frac, int):
                int_sum += frac
                n_samples.append(frac)
                if method is None:
                    method = 'number'
                assert method == 'number', "Split supports either numeric splits or lists of indices but not both"
            elif isinstance(frac, Iterable):
                if method is None:
                    method = 'indices'
                assert method == 'indices', "Split supports either numeric splits or lists of indices but not both"
            else:
                raise ValueError(
                    "split only accepts float, int, or iter[int] type splits, but {} was given".format(frac))
        assert frac_sum < 1, "total split fraction should sum to less than 1.0, but got: {}".format(frac_sum)
        assert int_sum < original_size, \
            "total split requirements ({}) should sum to less than dataset size ({})".format(int_sum, original_size)

        if method == 'number':
            if stratify is not None:
                splits = self._get_stratified_splits(n_samples, seed, stratify)
            else:
                splits = self._get_fractional_splits(n_samples, seed)
        else:  # method == 'indices':
            assert stratify is None, "Stratify may only be specified when splitting by count or fraction, not by index"
            splits = fractions
        splits = self._do_split(splits)
        FEDataset.fix_split_traceabilty(self, splits, fractions, seed, stratify)
        if len(fractions) == 1:
            return splits[0]
        return splits

    def _get_stratified_splits(self, split_counts: List[int], seed: Optional[int],
                               stratify: str) -> Sequence[Iterable[int]]:
        """Get sequence(s) of indices to split from the current dataset in order to generate new dataset(s).

        Args:
            split_counts: How many datapoints to include in each split.
            seed: What random seed, if any, to use when generating the split(s).
            stratify: A class key within the dataset with which to stratify the split (to approximately maintain class
                balance ratios before and after a split).

        Returns:
            Which data indices to include in each split of data. len(return[i]) == split_counts[i].
        """
        splits = []
        original_size = self._split_length()
        seed_offset = 0

        # Compute the distribution over the stratify key
        distribution = defaultdict(list)
        for idx in range(original_size):
            sample = self[idx]
            key = sample[stratify]
            if hasattr(key, "tobytes"):
                key = key.tobytes()  # Makes numpy arrays hashable
            distribution[key].append(idx)

        supply = {key: len(values) for key, values in distribution.items()}
        split_requests = [{key: (n_split * n_tot) / original_size
                           for key, n_tot in supply.items()} for n_split in split_counts]

        def transfer(source: Dict[Any, int], sink: Dict[Any, int], key: Any, request: int) -> int:
            allowance = min(request, source[key])
            source[key] -= allowance
            sink[key] += allowance
            return allowance

        # Sample splits proportional to the computed distribution
        for split_request, target in zip(split_requests, split_counts):
            split_actual = defaultdict(lambda: 0)
            total = 0
            # Step 1: Try to get as close to the target distribution as possible
            for key, request in split_request.items():
                request = 1 if 0 < request < 1 else round(request)  # Always want at least 1 sample from a class
                total += transfer(source=supply, sink=split_actual, key=key, request=request)
            # Step 2: Correct the error in the total count at the expense of an optimal distribution
            # |total - target| may be > n_keys due to rounding + supply shortage
            spare_last = True  # If we have drawn too many things, we will try not to reduce any classes beneath 1
            while total != target:
                old_total = total
                # Repeatedly add or shave 1 off of everything until we get the correct target number
                for key, requested in sorted(split_actual.items(), key=lambda x: x[1], reverse=True):
                    # reversed to start with the most abundant class
                    if total < target:
                        total += transfer(source=supply, sink=split_actual, key=key, request=1)
                    elif total > target and (not spare_last or requested > 1):
                        total -= transfer(source=split_actual, sink=supply, key=key, request=1)
                    if total == target:
                        break
                if old_total == total:
                    assert spare_last is True, "Cannot stratify the requested split. Please file a bug report."
                    # We weren't able to modify anything, so no choice but to reduce a 1-sample class to 0-sample
                    spare_last = False
            # Step 3: Perform the actual sampling
            split_indices = []
            for key, n_samples in split_actual.items():
                # Dicts have preserved insertion order since python 3.6, so we can increase the seed as we use it
                # to prevent any unintended patterns from emerging while still having consistency over multiple runs.
                # This wouldn't work if the order of encounter of a class can change (like in a generator), but in such
                # cases consistency would be impossible anyways.
                if seed is not None:
                    indices = random.Random(seed + seed_offset).sample(distribution[key], n_samples)
                    seed_offset += 1  # We'll use a different seed each time
                else:
                    indices = random.sample(distribution[key], n_samples)
                split_indices.extend(indices)
                # Sort to allow deterministic seed to work alongside sets
                distribution[key] = sorted(list(set(distribution[key]) - set(indices)))
            if seed is not None:
                random.Random(seed + seed_offset).shuffle(split_indices)
                seed_offset += 1
            else:
                random.shuffle(split_indices)
            splits.append(split_indices)
        return splits

    def _get_fractional_splits(self, split_counts: List[int], seed: Optional[int]) -> Sequence[Iterable[int]]:
        """Get sequence(s) of indices to split from the current dataset in order to generate new dataset(s).

        Args:
            split_counts: How many datapoints to include in each split.
            seed: What random seed, if any, to use when generating the split(s).

        Returns:
            Which data indices to include in each split of data. len(return[i]) == split_counts[i].
        """
        splits = []
        original_size = self._split_length()
        int_sum = sum(split_counts)
        # TODO - convert to a linear congruential generator for large datasets?
        # https://stackoverflow.com/questions/9755538/how-do-i-create-a-list-of-random-numbers-without-duplicates
        if seed is not None:
            indices = random.Random(seed).sample(range(original_size), int_sum)
        else:
            indices = random.sample(range(original_size), int_sum)
        start = 0
        for stop in split_counts:
            splits.append((indices[i] for i in range(start, start + stop)))
            start += stop
        return splits

    def _split_length(self) -> int:
        """The length of a dataset to be used for the purpose of computing splits.

        Useful if sub-classes want to split by something other than indices (see SiameseDirDataset for example).

        Returns:
            The apparent length of the dataset for the purpose of the .split() function
        """
        return len(self)

    def _do_split(self, splits: Sequence[Iterable[int]]) -> List['FEDataset']:
        """Split the current dataset apart into several smaller datasets.

        Args:
            splits: Which indices to remove from the current dataset in order to create new dataset(s). One dataset will
                be generated for every iterable within the `splits` sequence.

        Returns:
            New datasets generated by removing data at the indices specified by `splits` from the current dataset.
        """
        raise NotImplementedError

    def summary(self) -> DatasetSummary:
        """Generate a summary representation of this dataset.
        Returns:
            A summary representation of this dataset.
        """
        raise NotImplementedError

    def __str__(self):
        return str(self.summary())


@traceable(blacklist=('data', 'summary'))
class InMemoryDataset(FEDataset):
    """A dataset abstraction to simplify the implementation of datasets which hold their data in memory.

    Args:
        data: A dictionary like {data_index: {<instance dictionary>}}.
    """
    data: Dict[int, Dict[str, Any]]  # Index-based data dictionary

    def __init__(self, data: Dict[int, Dict[str, Any]]) -> None:
        self.data = data
        # Normally lru cache annotation is shared over all class instances, so calling cache_clear would reset all
        # caches (for example when calling .split()). Instead we make the lru cache per-instance
        self.summary = lru_cache(maxsize=1)(self.summary)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, index: Union[int, str]) -> Union[Dict[str, Any], np.ndarray, List[Any]]:
        """Look up data from the dataset.

        ```python
        data = fe.dataset.InMemoryDataset(...)  # {"x": <100>}, len(data) == 1000
        element = data[0]  # {"x": <100>}
        column = data["x"]  # <1000x100>
        ```

        Args:
            index: Either an int corresponding to a particular element of data, or a string in which case the
                corresponding column of data will be returned.

        Returns:
            A data dictionary if the index was an int, otherwise a column of data in list format.
        """
        if isinstance(index, int):
            return self.data[index]
        else:
            result = [elem[index] for elem in self.data.values()]
            if isinstance(result[0], np.ndarray):
                return np.array(result)
            return result

    def __setitem__(self, key: Union[int, str], value: Union[Dict[str, Any], Sequence[Any]]) -> None:
        """Modify data in the dataset.

        ```python
        data = fe.dataset.InMemoryDataset(...)  # {"x": <100>}, len(data) == 1000
        column = data["x"]  # <1000x100>
        column = column - np.mean(column)
        data["x"] = column
        ```

        Args:
            key: Either an int corresponding to a particular element of data, or a string in which case the
                corresponding column of data will be updated.
            value: The value to be inserted for the given `key`. Must be a dictionary if `key` is an integer. Otherwise
                must be a sequence with the same length as the current length of the dataset.

        Raises:
            AssertionError: If the `value` is inappropriate given the type of the `key`.
        """
        if isinstance(key, int):
            assert isinstance(value, Dict), "if setting a value using an integer index, must provide a dictionary"
            self.data[key] = value
        else:
            assert len(value) == len(self.data), \
                "input value must be of length {}, but had length {}".format(len(self.data), len(value))
            for i in range(len(self.data)):
                self.data[i][key] = value[i]
        self.summary.cache_clear()

    def _skip_init(self, data: Dict[int, Dict[str, Any]], **kwargs) -> 'InMemoryDataset':
        """A helper method to create new dataset instances without invoking their __init__ methods.

        Args:
            data: The data dictionary to be used in the new dataset.
            **kwargs: Any other member variables to be assigned in the new dataset.

        Returns:
            A new dataset based on the given inputs.
        """
        obj = self.__class__.__new__(self.__class__)
        obj.data = data
        for k, v in kwargs.items():
            if k == 'summary':
                continue  # Ignore summary object since we're going to re-initialize it
            else:
                obj.__setattr__(k, v)
        obj.summary = lru_cache(maxsize=1)(obj.summary)
        return obj

    def _do_split(self, splits: Sequence[Iterable[int]]) -> List['InMemoryDataset']:
        """Split the current dataset apart into several smaller datasets.

        Args:
            splits: Which indices to remove from the current dataset in order to create new dataset(s). One dataset will
                be generated for every iterable within the `splits` sequence.

        Returns:
            New Datasets generated by removing data at the indices specified by `splits` from the current dataset.
        """
        results = []
        for split in splits:
            data = {new_idx: self.data.pop(old_idx) for new_idx, old_idx in enumerate(split)}
            results.append(self._skip_init(data, **{k: v for k, v in self.__dict__.items() if k not in {'data'}}))
        # Re-key the remaining data to be contiguous from 0 to new max index
        self.data = {new_idx: v for new_idx, (old_idx, v) in enumerate(self.data.items())}
        self.summary.cache_clear()
        return results

    def summary(self) -> DatasetSummary:
        """Generate a summary representation of this dataset.
        Returns:
            A summary representation of this dataset.
        """
        # We will check whether the dataset is doing additional pre-processing on top of the self.data keys. If not we
        # can extract extra information about the data without incurring a large computational time cost
        final_example = self[0]
        original_example = self.data[0]
        keys = final_example.keys()
        shapes = {}
        dtypes = {}
        n_unique_vals = defaultdict(lambda: 0)
        for key in keys:
            final_val = final_example[key]
            # TODO - if val is empty list, should find a sample which has entries
            dtypes[key] = get_type(final_val)
            shapes[key] = get_shape(final_val)

            # Check whether type and shape have changed by get_item
            if key in original_example:
                original_val = original_example[key]
                original_dtype = get_type(original_val)
                original_shape = get_shape(original_val)

                # If no changes, then we can relatively quickly count the unique values using self.data
                if dtypes[key] == original_dtype and shapes[key] == original_shape and isinstance(
                        original_val, Hashable):
                    n_unique_vals[key] = len({self.data[i][key] for i in range(len(self.data))})

        key_summary = {
            key: KeySummary(dtype=dtypes[key], num_unique_values=n_unique_vals[key] or None, shape=shapes[key])
            for key in keys
        }
        return DatasetSummary(num_instances=len(self), keys=key_summary)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import shutil
import tempfile
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.op.tensorop import TensorOp
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.summary import Summary
from fastestimator.test.unittest_util import OneLayerTorchModel, is_equal, one_layer_tf_model, sample_system_object, \
    sample_system_object_torch
from fastestimator.trace.trace import Trace


def get_model_names(system):
    model_names = []
    for model in system.network.models:
        model_names.append(model.model_name)
    return model_names


class TestTensorOp(TensorOp):
    def __init__(self, inputs, outputs, mode, var1):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.var1 = var1


class TestTrace(Trace):
    def __init__(self, var1):
        super().__init__()
        self.var1 = var1


def test_model(submodel):
    inp = tf.keras.layers.Input([3])
    x = submodel(inp)
    x = x + 1
    model = tf.keras.models.Model(inputs=inp, outputs=x)
    return model


class TestModel(torch.nn.Module):
    def __init__(self, submodel):
        super().__init__()
        self.submodel = submodel

    def forward(self, x):
        x = self.submodel(x)
        x = x + 1
        return x


class TestSystem(unittest.TestCase):
    def test_save_and_load_state_torch(self):
        """ `save_state` and `load_state` of an entire system are highly dependent on the implementation of __setstate__
        and __getstate__ of ops, traces, datasets ... etc. The save_state and load_state function should be tested in
        the testing script of that perticular component
        (ex: The test of save_state and load_state with EpochScheduler is located in
             PR_test/integration_test/schedule/test_epoch_scheduler.py)
        """
        system = sample_system_object_torch()
        model_names = get_model_names(system)
        global_step = 100
        epoch_idx = 10
        save_path = tempfile.mkdtemp()

        system.global_step = global_step
        system.epoch_idx = epoch_idx
        with self.subTest("Check state files were created"):
            system.save_state(save_dir=save_path)
            self.assertTrue(os.path.exists(os.path.join(save_path, 'objects.pkl')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'system.json')))
            for model_name in model_names:
                self.assertTrue(os.path.exists(os.path.join(save_path, f'{model_name}.pt')))
                self.assertTrue(os.path.exists(os.path.join(save_path, f'{model_name}_opt.pt')))

        system = sample_system_object_torch()
        with self.subTest("Check that state loads properly"):
            system.load_state(save_path)
            self.assertEqual(global_step, system.global_step)
            self.assertEqual(epoch_idx, system.epoch_idx)

        if os.path.exists(save_path):
            shutil.rmtree(save_path)

    def test_save_and_load_state_tf(self):
        """ `save_state` and `load_state` of an entire system are highly dependent on the implementation of __setstate__
        and __getstate__ of ops, traces, datasets ... etc. The save_state and load_state function should be tested in
        the testing script of that particular component
        (ex: The test of save_state and load_state with EpochScheduler is located in
             PR_test/integration_test/schedule/test_epoch_scheduler.py)
        """
        system = sample_system_object()
        model_names = get_model_names(system)
        global_step = 100
        epoch_idx = 10
        save_path = tempfile.mkdtemp()

        system.global_step = global_step
        system.epoch_idx = epoch_idx
        with self.subTest("Check state files were created"):
            system.save_state(save_dir=save_path)
            self.assertTrue(os.path.exists(os.path.join(save_path, 'objects.pkl')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'system.json')))
            for model_name in model_names:
                self.assertTrue(os.path.exists(os.path.join(save_path, f'{model_name}.h5')))
                self.assertTrue(os.path.exists(os.path.join(save_path, f'{model_name}_opt.pkl')))

        system = sample_system_object()
        with self.subTest("Check that state loads properly"):
            system.load_state(save_path)
            self.assertEqual(global_step, system.global_step)
            self.assertEqual(epoch_idx, system.epoch_idx)

        if os.path.exists(save_path):
            shutil.rmtree(save_path)

    def test_shared_variables_within_traces(self):
        save_path = tempfile.mkdtemp()

        system = sample_system_object()
        shared_trace_var = tf.Variable(initial_value=2, trainable=True)
        system.traces.append(TestTrace(shared_trace_var))
        system.traces.append(TestTrace(shared_trace_var))
        shared_trace_var.assign_add(1)

        system.save_state(save_dir=save_path)

        # Re-initialize
        system = sample_system_object()
        shared_trace_var = tf.Variable(initial_value=2, trainable=True)
        system.traces.append(TestTrace(shared_trace_var))
        system.traces.append(TestTrace(shared_trace_var))

        system.load_state(load_dir=save_path)

        with self.subTest("Check variable value was re-loaded"):
            self.assertEqual(3, system.traces[-1].var1.numpy())
            self.assertEqual(3, system.traces[-2].var1.numpy())

        with self.subTest("Check that variable is still shared"):
            system.traces[-1].var1.assign(5)
            self.assertEqual(5, system.traces[-1].var1.numpy())
            self.assertEqual(5, system.traces[-2].var1.numpy())

        with self.subTest("Check that variable is still linked to outside code"):
            shared_trace_var.assign(7)
            self.assertEqual(7, system.traces[-1].var1.numpy())
            self.assertEqual(7, system.traces[-2].var1.numpy())

    def test_shared_variables_over_object_types(self):
        save_path = tempfile.mkdtemp()

        system = sample_system_object()
        shared_var = tf.Variable(initial_value=2, trainable=True)
        system.traces.append(TestTrace(shared_var))
        system.network.ops[0].fe_test_var_1 = shared_var
        shared_var.assign_add(1)

        system.save_state(save_dir=save_path)

        # Re-initialize
        system = sample_system_object()
        shared_var = tf.Variable(initial_value=2, trainable=True)
        system.traces.append(TestTrace(shared_var))
        system.network.ops[0].fe_test_var_1 = shared_var

        system.load_state(load_dir=save_path)

        with self.subTest("Check variable value was re-loaded"):
            self.assertEqual(3, system.traces[-1].var1.numpy())
            self.assertEqual(3, system.network.ops[0].fe_test_var_1.numpy())

        with self.subTest("Check that variable is still shared"):
            system.traces[-1].var1.assign(5)
            self.assertEqual(5, system.traces[-1].var1.numpy())
            self.assertEqual(5, system.network.ops[0].fe_test_var_1.numpy())

        with self.subTest("Check that variable is still linked to outside code"):
            shared_var.assign(7)
            self.assertEqual(7, system.traces[-1].var1.numpy())
            self.assertEqual(7, system.network.ops[0].fe_test_var_1.numpy())

    def test_shared_variable_over_model_tf(self):
        def instantiate_system():
            system = sample_system_object()
            submodel = one_layer_tf_model()
            model = fe.build(model_fn=lambda: test_model(submodel), optimizer_fn='adam', model_name='tf')
            model2 = fe.build(model_fn=lambda: test_model(submodel), optimizer_fn='adam', model_name='tf2')
            system.network = fe.Network(ops=[
                ModelOp(model=model, inputs="x_out", outputs="y_pred"),
                ModelOp(model=model2, inputs="x_out", outputs="y_pred2"),
            ])

            return system

        system = instantiate_system()

        # make some change
        new_weight = [np.array([[1.0], [1.0], [1.0]])]
        system.network.ops[0].model.layers[1].set_weights(new_weight)

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        shared_variable = system.network.ops[0].model.layers[1]
        system.load_state(save_path)

        with self.subTest("Check model varaible was re-loaded"):
            self.assertTrue(is_equal(new_weight, system.network.ops[0].model.layers[1].get_weights()))
            self.assertTrue(is_equal(new_weight, system.network.ops[1].model.layers[1].get_weights()))

        with self.subTest("Check model variable is still shared"):
            new_weight = [np.array([[2.0], [2.0], [2.0]])]
            system.network.ops[0].model.layers[1].set_weights(new_weight)
            self.assertTrue(
                is_equal(system.network.ops[0].model.layers[1].get_weights(),
                         system.network.ops[1].model.layers[1].get_weights()))

        with self.subTest("Check that variable is still linked to outside code"):
            new_weight = [np.array([[3.0], [3.0], [3.0]])]
            shared_variable.set_weights(new_weight)
            self.assertTrue(is_equal(new_weight, system.network.ops[0].model.layers[1].get_weights()))

    def test_shared_variable_over_model_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            submodel = OneLayerTorchModel()
            model = fe.build(model_fn=lambda: TestModel(submodel), optimizer_fn='adam', model_name='torch')
            model2 = fe.build(model_fn=lambda: TestModel(submodel), optimizer_fn='adam', model_name='torch2')
            system.network = fe.Network(ops=[
                ModelOp(model=model, inputs="x_out", outputs="y_pred"),
                ModelOp(model=model2, inputs="x_out", outputs="y_pred2"),
            ])

            return system

        system = instantiate_system()

        # check if multi-gpu system
        if torch.cuda.device_count() > 1:
            model_0 = system.network.ops[0].model.module
            model_1 = system.network.ops[1].model.module
        else:
            model_0 = system.network.ops[0].model
            model_1 = system.network.ops[1].model

        # make some change
        new_weight = torch.tensor([[1, 1, 1]], dtype=torch.float32)
        model_0.submodel.fc1.weight.data = new_weight

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        shared_variable = model_0.submodel.fc1.weight
        system.load_state(save_path)

        with self.subTest("Check model varaible was re-loaded"):
            self.assertTrue(is_equal(new_weight, model_0.submodel.fc1.weight.data))
            self.assertTrue(is_equal(new_weight, model_1.submodel.fc1.weight.data))

        with self.subTest("Check model variable is still shared"):
            new_weight = torch.tensor([[2, 2, 2]], dtype=torch.float32)
            model_0.submodel.fc1.weight.data = new_weight
            self.assertTrue(is_equal(model_0.submodel.fc1.weight.data, model_1.submodel.fc1.weight.data))

        with self.subTest("Check that variable is still linked to outside code"):
            new_weight = torch.tensor([[3, 3, 3]], dtype=torch.float32)
            shared_variable.data = new_weight
            self.assertTrue(is_equal(new_weight, model_0.submodel.fc1.weight.data))

    def test_shared_tf_variable_among_top_trace(self):
        def instantiate_system():
            system = sample_system_object()
            model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam', model_name='tf')
            var1 = tf.Variable(initial_value=1, trainable=True)
            system.network = fe.Network(ops=[
                TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var1=var1),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            system.traces.append(TestTrace(var1=var1))

            return system

        system = instantiate_system()

        # make some change
        var1_new_val = 2
        system.traces[0].var1.assign(var1_new_val)

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        var1 = system.traces[0].var1
        system.load_state(save_path)

        with self.subTest("Check both trace and tensorop variables are reloaded"):
            self.assertEqual(var1_new_val, system.traces[0].var1.numpy())
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

        with self.subTest("Check trace and tensorop variables are still shared"):
            var1_new_val = 3
            system.traces[0].var1.assign(var1_new_val)
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

        with self.subTest("Check that variable is still linked to outside code"):
            var1_new_val = 4
            var1.assign(var1_new_val)
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

    def test_shared_torch_variable_among_top_trace(self):
        def instantiate_system():
            system = sample_system_object_torch()
            model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='torch')
            var1 = torch.tensor(1.0)
            system.network = fe.Network(ops=[
                TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var1=var1),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            system.traces.append(TestTrace(var1=var1))

            return system

        system = instantiate_system()

        # make some change
        var1_new_val = 2.0
        system.traces[0].var1.copy_(torch.tensor(var1_new_val))

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        var1 = system.traces[0].var1
        system.load_state(save_path)

        with self.subTest("Check both trace and tensorop variables are reloaded"):
            self.assertEqual(var1_new_val, system.traces[0].var1.numpy())
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

        with self.subTest("Check trace and tensorop variables are still shared"):
            var1_new_val = 3.0
            system.traces[0].var1.copy_(torch.tensor(var1_new_val))
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

        with self.subTest("Check that variable is still linked to outside code"):
            var1_new_val = 4.0
            var1.copy_(torch.tensor(var1_new_val))
            self.assertEqual(var1_new_val, system.network.ops[0].var1.numpy())

    def test_save_and_load_custom_graphs(self):
        def instantiate_system():
            system = sample_system_object()
            return system

        # Make some data
        example_graph1 = Summary(name="exp1")
        example_graph1.history['train']['acc'][0] = 0.95
        example_graph1.history['train']['acc'][10] = 0.87

        example_graph2 = Summary(name="class1")
        example_graph2.history['eval']['dice'][15] = 0.34
        example_graph2.history['eval']['dice'][25] = 0.75

        example_graph3 = Summary(name="class2")
        example_graph3.history['eval']['dice'][15] = 0.21
        example_graph3.history['eval']['dice'][25] = 0.93

        system = instantiate_system()
        system.add_graph("sample_single_graph", example_graph1)
        system.add_graph("sample_list_graph", [example_graph2, example_graph3])

        # Save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        # Check the results
        with self.subTest('Check that 2 graphs are recovered'):
            self.assertEqual(2, len(system.custom_graphs))
        with self.subTest('Check that the single graph is recovered'):
            self.assertIn("sample_single_graph", system.custom_graphs)
            self.assertEqual('exp1', system.custom_graphs['sample_single_graph'][0].name)
            self.assertEqual(0.87, system.custom_graphs['sample_single_graph'][0].history['train']['acc'][10])
        with self.subTest('Check that multi-graphs are recovered'):
            self.assertIn("sample_list_graph", system.custom_graphs)
            self.assertEqual('class1', system.custom_graphs['sample_list_graph'][0].name)
            self.assertEqual('class2', system.custom_graphs['sample_list_graph'][1].name)
            self.assertEqual(0.34, system.custom_graphs['sample_list_graph'][0].history['eval']['dice'][15])
            self.assertEqual(0.21, system.custom_graphs['sample_list_graph'][1].history['eval']['dice'][15])



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf

from fastestimator.architecture.tensorflow import AttentionUNet
from fastestimator.architecture.tensorflow.attention_unet import _check_input_size


class TestAttentionUNet(unittest.TestCase):
    def test_attention_unet_default(self):
        data = np.ones((1, 128, 128, 1))
        input_data = tf.constant(data)
        attention_unet = AttentionUNet()
        output_shape = attention_unet(input_data).numpy().shape
        self.assertEqual(output_shape, (1, 128, 128, 1))

    def test_attention_unet_specific_input_size(self):
        size = (16, 16, 1)
        data = np.ones((1, ) + size)
        input_data = tf.constant(data)
        attention_unet = AttentionUNet(input_size=size)
        output_shape = attention_unet(input_data).numpy().shape
        self.assertEqual(output_shape, (1, ) + size)


class TestCheckInputSize(unittest.TestCase):
    def test_check_input_size(self):
        with self.subTest("length not 3"):
            with self.assertRaises(ValueError):
                _check_input_size((1, ))

        with self.subTest("width or height is not a multiple of 16"):
            with self.assertRaises(ValueError):
                _check_input_size((18, 16, 1))

            with self.assertRaises(ValueError):
                _check_input_size((32, 100, 1))

            with self.assertRaises(ValueError):
                _check_input_size((48, 0, 1))

        with self.subTest("both are multiples of 16"):
            _check_input_size((16, 48, 1))
            _check_input_size((128, 64, 3))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
import inspect
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.network import BaseNetwork
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class Repeat(TensorOp):
    """Repeat a TensorOp several times in a row.

    Args:
        op: A TensorOp to be run one or more times in a row.
        repeat: How many times to repeat the `op`. This can also be a function return, in which case the function input
            names will be matched to keys in the data dictionary, and the `op` will be repeated until the function
            evaluates to False. The function evaluation will happen at the end of a forward call, so the `op` will
            always be evaluated at least once. If a function is provided, any TF ops which are wrapped by Repeat will
            not have access to the gradient tape, nor to previously deferred model update functions.
        max_iter: A limit to how many iterations will be run (or None for no limit).

    Raises:
        ValueError: If `repeat`, `op`, or max_iter are invalid.
    """
    def __init__(self, op: TensorOp, repeat: Union[int, Callable[..., bool]] = 1,
                 max_iter: Optional[int] = None) -> None:
        self.repeat_inputs = []
        extra_reqs = []
        if max_iter is None:
            self.max_iter = max_iter
        else:
            if max_iter < 1:
                raise ValueError(f"Repeat requires max_iter to be >=1, but got {max_iter}")
            self.max_iter = max_iter - 1  # -1 b/c the first invocation happens outside the while loop
        if isinstance(repeat, int):
            if repeat < 1:
                raise ValueError(f"Repeat requires repeat to be >= 1, but got {repeat}")
            if max_iter:
                raise ValueError("Do not set max_iter when repeat is an integer")
        else:
            self.repeat_inputs.extend(inspect.signature(repeat).parameters.keys())
            extra_reqs = list(set(self.repeat_inputs) - set(op.outputs))
        self.repeat = repeat
        super().__init__(inputs=op.inputs + extra_reqs, outputs=op.outputs, mode=op.mode, ds_id=op.ds_id)
        self.ops = [op]
        self.retain_graph = None
        self.while_fn = None

    @property
    def op(self) -> TensorOp:
        return self.ops[0]

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        self.op.build(framework, device)
        if framework == 'tf':
            self.while_fn = self._tf_while
        else:
            self.while_fn = self._torch_while

    def get_fe_models(self) -> Set[Model]:
        return self.op.get_fe_models()

    def get_fe_loss_keys(self) -> Set[str]:
        return self.op.get_fe_loss_keys()

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        if retain is not None:
            self.retain_graph = retain
        return self.op.fe_retain_graph(retain)

    def __getstate__(self) -> Dict[str, List[Dict[Any, Any]]]:
        return {'ops': [elem.__getstate__() if hasattr(elem, '__getstate__') else {} for elem in self.ops]}

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> List[Tensor]:
        # Set retain to true since might loop over a gradient aware op
        self.op.fe_retain_graph(True)

        data = {key: elem for key, elem in zip(self.inputs, data)}
        if isinstance(self.repeat, int):
            for i in range(self.repeat - 1):
                # Perform n-1 rounds with all ops having retain_graph == True
                BaseNetwork._forward_batch(data, state, self.ops)
            # Let retain be whatever it was meant to be for the final sequence
            self.op.fe_retain_graph(self.retain_graph)
            # Final round of ops
            BaseNetwork._forward_batch(data, state, self.ops)
        else:
            BaseNetwork._forward_batch(data, state, self.ops)
            data = self.while_fn(data, state)
            # TODO - Find some magic way to invoke this at the right moment
            self.op.fe_retain_graph(self.retain_graph)
        return [data[key] for key in self.outputs]

    def _torch_while(self, data: Dict[str, Tensor], state: Dict[str, Any]) -> Dict[str, Tensor]:
        """A helper function to invoke a while loop.

        Args:
            data: A data dictionary to be used during looping.
            state: The state variables to be considered during looping.

        Returns:
            A reference to the updated data dictionary.
        """
        i = 0
        while self.repeat(*[data[var_name] for var_name in self.repeat_inputs]):
            if self.max_iter and i >= self.max_iter:
                break
            BaseNetwork._forward_batch(data, state, self.ops)
            i += 1
        return data

    def _tf_while(self, data: Dict[str, Tensor], state: Dict[str, Any]) -> Dict[str, Tensor]:
        """A helper function to invoke a while loop.

        Args:
            data: A data dictionary to be used during looping.
            state: The state variables to be considered during looping.

        Returns:
            A reference to the updated data dictionary.
        """
        args = ([data[var_name] for var_name in self.repeat_inputs], data)
        # Use functools.partial since state may contain objects which cannot be cast to tensors (ex. gradient tape)
        args = tf.while_loop(self._tf_cond,
                             functools.partial(self._tf_body, state=state),
                             args,
                             maximum_iterations=self.max_iter)
        return args[1]

    def _tf_cond(self, cnd: List[Tensor], data: Dict[str, Tensor]) -> bool:
        """A helper function determine whether to keep invoking the while method.

        Note that `data` and `state` are unused here, but required since tf.while_loop needs the cond and body to have
        the same input argument signatures.

        Args:
            cnd: A list of arguments to be passed to the condition function.
            data: A data dictionary to be used during looping.

        Returns:
            Whether to continue looping.
        """
        return self.repeat(*cnd)

    def _tf_body(self, cnd: List[Tensor], data: Dict[str, Tensor],
                 state: Dict[str, Any]) -> Tuple[List[Tensor], Dict[str, Tensor]]:
        """A helper function to execute the body of a while method.

        Note that `cnd` is unused here, but required since tf.while_loop needs the cond and body to have the same input
        argument signatures.

        Args:
            cnd: A list of arguments to be passed to the condition function.
            data: A data dictionary to be used during looping.
            state: The state variables to be considered during looping.

        Returns:
            The updated `cnd` values, along with the modified data and state dictionaries.
        """
        BaseNetwork._forward_batch(data, state, self.ops)
        return [data[var_name] for var_name in self.repeat_inputs], data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.schedule import EpochScheduler, RepeatScheduler, get_current_items, get_signature_epochs


class TestSchedule(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.scheduler = RepeatScheduler(['a', 'b', 'c', 'c'])
        cls.epoch_scheduler = EpochScheduler({1: 'a', 2: 'b', 30: None})
        cls.actual_current_items = ['a', 'b', 'c', 'c']
        cls.signature_epochs = [1, 2, 3, 5, 30, 31, 33]

    def test_get_current_items(self):
        current_items = get_current_items(items=[self.scheduler], run_modes="train", epoch=None)
        self.assertEqual(current_items, self.actual_current_items)

    def test_get_current_item_epoch(self):
        current_items = get_current_items(items=[self.scheduler], run_modes="train", epoch=2)
        self.assertEqual(current_items, ['b'])

    def test_get_signature_epochs(self):
        epochs = get_signature_epochs([self.scheduler, self.epoch_scheduler], total_epochs=50, mode="train")
        self.assertEqual(epochs, self.signature_epochs)



#  Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import os
from typing import Optional, Set, Union, Iterable

import calibration as cal
import dill
import numpy as np

from fastestimator.summary.system import System
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list, to_number


@traceable()
class PBMCalibrator(Trace):
    """A trace to generate a PlattBinnerMarginalCalibrator given a set of predictions.

    Unlike many common calibration error correction algorithms, this one has actual theoretical bounds on the quality
    of its output: https://arxiv.org/pdf/1909.10155v1.pdf. This trace is commonly used together with the Calibrate
    NumpyOp for postprocessing. This trace will collect data from whichever `mode` it is set to run on in order to
    perform empirical probability calibration. The calibrated predictions will be output on epoch end. The trained
    calibration function will also be saved if `save_path` is provided.

    Args:
        true_key: Name of the key that corresponds to ground truth in the batch dictionary.
        pred_key: Name of the key that corresponds to predicted score in the batch dictionary.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        output_name: What to call the output from this trace. If None, the default will be '<pred_key>_calibrated'.
        save_path: Where to save the calibrator generated by this Trace. If None, then no saving will be performed.
        save_if_key: Name of a key to control whether to save the calibrator. For example "since_best_acc". If provided,
            then the calibrator will only be saved when the save_if_key value is zero.

    Raises:
        ValueError: If 'save_if_key' is provided but no 'save_path' is given.
    """
    system: System

    def __init__(self,
                 true_key: str,
                 pred_key: str,
                 output_name: Optional[str] = None,
                 save_path: Optional[str] = None,
                 save_if_key: Optional[str] = None,
                 mode: Union[str, Set[str]] = "eval",
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        if output_name is None:
            output_name = f"{pred_key}_calibrated"
        if save_if_key is not None and save_path is None:
            raise ValueError("If 'save_if_key' is provided, then a 'save_path' must also be provided.")
        super().__init__(inputs=[true_key, pred_key] + to_list(save_if_key),
                         outputs=output_name,
                         mode=mode,
                         ds_id=ds_id)
        self.y_true = []
        self.y_pred = []
        if save_path is not None:
            save_path = os.path.abspath(os.path.normpath(save_path))
        self.save_path = save_path

    @property
    def true_key(self) -> str:
        return self.inputs[0]

    @property
    def pred_key(self) -> str:
        return self.inputs[1]

    @property
    def save_key(self) -> Optional[str]:
        if len(self.inputs) == 3:
            return self.inputs[2]
        return None

    def on_epoch_begin(self, data: Data) -> None:
        self.y_true = []
        self.y_pred = []

    def on_batch_end(self, data: Data) -> None:
        y_true, y_pred = to_number(data[self.true_key]), to_number(data[self.pred_key])
        if y_true.shape[-1] > 1 and y_true.ndim > 1:
            y_true = np.argmax(y_true, axis=-1)
        assert y_pred.shape[0] == y_true.shape[0]
        self.y_true.extend(y_true)
        self.y_pred.extend(y_pred)

    def on_epoch_end(self, data: Data) -> None:
        self.y_true = np.squeeze(np.stack(self.y_true))
        self.y_pred = np.stack(self.y_pred)
        calibrator = cal.PlattBinnerMarginalCalibrator(num_calibration=len(self.y_true), num_bins=10)
        calibrator.train_calibration(probs=self.y_pred, labels=self.y_true)
        if self.save_path:
            if not self.save_key or (self.save_key and to_number(data[self.save_key]) == 0):
                with open(self.save_path, 'wb') as f:
                    dill.dump(calibrator.calibrate, file=f)
                print(f"FastEstimator-PBMCalibrator: Calibrator written to {self.save_path}")
        data.write_without_log(self.outputs[0], calibrator.calibrate(self.y_pred))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import pickle
from typing import Optional, Union

import tensorflow as tf
import tensorflow_addons as tfa
import torch

from fastestimator.backend.get_lr import get_lr


def save_model(model: Union[tf.keras.Model, torch.nn.Module],
               save_dir: str,
               model_name: Optional[str] = None,
               save_optimizer: bool = False,
               save_architecture: bool = False) -> str:
    """Save `model` weights to a specific directory.

    This method can be used with TensorFlow models:
    ```python
    m = fe.build(fe.architecture.tensorflow.LeNet, optimizer_fn="adam")
    fe.backend.save_model(m, save_dir="/tmp", model_name="test")  # Generates 'test.h5' file inside /tmp directory
    ```

    This method can be used with PyTorch models:
    ```python
    m = fe.build(fe.architecture.pytorch.LeNet, optimizer_fn="adam")
    fe.backend.save_model(m, save_dir="/tmp", model_name="test")  # Generates 'test.pt' file inside /tmp directory
    ```

    Args:
        model: A neural network instance to save.
        save_dir: Directory into which to write the `model` weights.
        model_name: The name of the model (used for naming the weights file). If None, model.model_name will be used.
        save_optimizer: Whether to save optimizer. If True, optimizer will be saved in a separate file at same folder.
        save_architecture: Whether to also save the entire model architecture so that the model can be reloaded without
            needing access to the code which generated it. This option is only available for TensorFlow models.

    Returns:
        The saved model path.

    Raises:
        ValueError: If `model` is an unacceptable data type, of if a user tries to save architecture of a PyTorch model.
    """
    assert hasattr(model, "fe_compiled") and model.fe_compiled, "model must be built by fe.build"
    if model_name is None:
        model_name = model.model_name
    save_dir = os.path.normpath(save_dir)
    os.makedirs(save_dir, exist_ok=True)
    if isinstance(model, tf.keras.Model):
        model_path = os.path.join(save_dir, "{}.h5".format(model_name))
        model.save_weights(model_path)
        if save_architecture:
            model.save(filepath=os.path.join(save_dir, model_name), include_optimizer=save_optimizer)
        if save_optimizer:
            assert model.current_optimizer, "optimizer does not exist"
            optimizer_path = os.path.join(save_dir, "{}_opt.pkl".format(model_name))
            with open(optimizer_path, 'wb') as f:
                saved_data = {'weights': model.current_optimizer.get_weights(), 'lr': get_lr(model)}
                if isinstance(model.current_optimizer, tfa.optimizers.DecoupledWeightDecayExtension) or hasattr(
                        model.current_optimizer, "inner_optimizer") and isinstance(
                            model.current_optimizer.inner_optimizer, tfa.optimizers.DecoupledWeightDecayExtension):
                    saved_data['weight_decay'] = tf.keras.backend.get_value(model.current_optimizer.weight_decay)
                pickle.dump(saved_data, f)
        return model_path
    elif isinstance(model, torch.nn.Module):
        model_path = os.path.join(save_dir, "{}.pt".format(model_name))
        torch.save(model.state_dict(), model_path)
        if save_architecture:
            raise ValueError("Sorry, architecture saving is not currently enabled for PyTorch")
        if save_optimizer:
            assert model.current_optimizer, "optimizer does not exist"
            optimizer_path = os.path.join(save_dir, "{}_opt.pt".format(model_name))
            torch.save(model.current_optimizer.state_dict(), optimizer_path)
        return model_path
    else:
        raise ValueError("Unrecognized model instance {}".format(type(model)))



#!/usr/bin/env python
# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import sys

from fastestimator.cli.history import configure_history_parser
from fastestimator.cli.logs import configure_log_parser
from fastestimator.cli.run import configure_run_parser
from fastestimator.cli.train import configure_test_parser, configure_train_parser


def run_main(argv) -> None:
    """A function which invokes the various argument parsers and then runs the requested subroutine.
    """
    parser = argparse.ArgumentParser(allow_abbrev=False)
    subparsers = parser.add_subparsers()
    # In python 3.7 the following 2 lines could be put into the .add_subparsers() call
    subparsers.required = True
    subparsers.dest = 'mode'
    configure_train_parser(subparsers)
    configure_test_parser(subparsers)
    configure_run_parser(subparsers)
    configure_log_parser(subparsers)
    configure_history_parser(subparsers)
    args, unknown = parser.parse_known_args(argv)
    args.func(vars(args), unknown)


def main() -> None:
    run_main(sys.argv[1:])


if __name__ == '__main__':
    main()



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, ChainMap, Dict, List, MutableMapping, Optional


class Data(ChainMap[str, Any]):
    """A class which contains prediction and batch data.

    This class is intentionally not @traceable.

    Data objects can be interacted with as if they are regular dictionaries. They are however, actually a combination of
    two dictionaries, a dictionary for trace communication and a dictionary of prediction+batch data. In general, data
    written into the trace dictionary will be logged by the system, whereas data in the pred+batch dictionary will not.
    We therefore provide helper methods to write entries into `Data` which are intended or not intended for logging.

    ```python
    d = fe.util.Data({"a":0, "b":1, "c":2})
    a = d["a"]  # 0
    d.write_with_log("d", 3)
    d.write_without_log("e", 5)
    d.write_with_log("a", 4)
    a = d["a"]  # 4
    r = d.read_logs(extra_keys={"c"})  # {"c":2, "d":3, "a":4}
    ```

    Args:
        batch_data: The batch data dictionary. In practice this is itself often a ChainMap containing separate
            prediction and batch dictionaries.
    """
    maps: List[MutableMapping[str, Any]]

    def __init__(self, batch_data: Optional[MutableMapping[str, Any]] = None) -> None:
        super().__init__({}, batch_data or {})

    def write_with_log(self, key: str, value: Any) -> None:
        """Write a given `value` into the `Data` dictionary with the intent that it be logged.

        Args:
            key: The key to associate with the new entry.
            value: The new entry to be written.
        """
        self.__setitem__(key, value)

    def write_without_log(self, key: str, value: Any) -> None:
        """Write a given `value` into the `Data` dictionary with the intent that it not be logged.

        Args:
            key: The key to associate with the new entry.
            value: The new entry to be written.
        """
        self.maps[1][key] = value

    def read_logs(self) -> Dict[str, Any]:
        """Read all values from the `Data` dictionary which were intended to be logged.

        Returns:
            A dictionary of all of the keys and values to be logged.
        """
        return self.maps[0]



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, List, Optional, Tuple, Union

import cv2
from albumentations.augmentations.transforms import OpticalDistortion as OpticalDistortionAlb

from fastestimator.op.numpyop.multivariate.multivariate import MultiVariateAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class OpticalDistortion(MultiVariateAlbumentation):
    """Apply optical distortion to an image / mask.

    Args:
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        image_in: The key of an image to be modified.
        mask_in: The key of a mask to be modified (with the same random factors as the image).
        masks_in: The key of masks to be modified (with the same random factors as the image).
        image_out: The key to write the modified image (defaults to `image_in` if None).
        mask_out: The key to write the modified mask (defaults to `mask_in` if None).
        masks_out: The key to write the modified masks (defaults to `masks_in` if None).
        distort_limit: If distort_limit is a single float, the range will be (-distort_limit, distort_limit).
        shift_limit: If shift_limit is a single float, the range will be (-shift_limit, shift_limit).
        interpolation: Flag that is used to specify the interpolation algorithm. Should be one of:
            cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA, cv2.INTER_LANCZOS4.
        border_mode: Flag that is used to specify the pixel extrapolation method. Should be one of:
            cv2.BORDER_CONSTANT, cv2.BORDER_REPLICATE, cv2.BORDER_REFLECT, cv2.BORDER_WRAP, cv2.BORDER_REFLECT_101.
        value: Padding value if border_mode is cv2.BORDER_CONSTANT.
        mask_value: Padding value if border_mode is cv2.BORDER_CONSTANT applied for masks.

    Image types:
        uint8, float32
    """
    def __init__(self,
                 distort_limit: Union[float, Tuple[float, float]] = 0.05,
                 shift_limit: Union[float, Tuple[float, float]] = 0.05,
                 interpolation: int = cv2.INTER_LINEAR,
                 border_mode: int = cv2.BORDER_REFLECT_101,
                 value: Union[None, int, float, List[int], List[float]] = None,
                 mask_value: Union[None, int, float, List[int], List[float]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 image_in: Optional[str] = None,
                 mask_in: Optional[str] = None,
                 masks_in: Optional[str] = None,
                 image_out: Optional[str] = None,
                 mask_out: Optional[str] = None,
                 masks_out: Optional[str] = None):
        super().__init__(
            OpticalDistortionAlb(distort_limit=distort_limit,
                                 shift_limit=shift_limit,
                                 interpolation=interpolation,
                                 border_mode=border_mode,
                                 value=value,
                                 mask_value=mask_value,
                                 always_apply=True),
            image_in=image_in,
            mask_in=mask_in,
            masks_in=masks_in,
            bbox_in=None,
            keypoints_in=None,
            image_out=image_out,
            mask_out=mask_out,
            masks_out=masks_out,
            bbox_out=None,
            keypoints_out=None,
            bbox_params=None,
            keypoint_params=None,
            mode=mode,
            ds_id=ds_id)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import random
from typing import Any, Dict, Iterable, List, Union

import numpy as np
from PIL import Image, ImageEnhance

from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import param_to_range


@traceable()
class Brightness(NumpyOp):
    """Randomly change the brightness of an image.

    This is a wrapper for functionality provided by the PIL library:
    https://github.com/python-pillow/Pillow/tree/master/src/PIL.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        limit: Factor range for changing brightness. If limit is a single float, the range will be (-limit, limit).
            A factor of 0.0 gives a black image and a factor of 1.0 gives the original image.

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 limit: float = 0.54):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.limit = param_to_range(limit)
        self.in_list, self.out_list = True, True

    def set_rua_level(self, magnitude_coef: float) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """
        param_mid = (self.limit[1] + self.limit[0]) / 2
        param_extent = magnitude_coef * ((self.limit[1] - self.limit[0]) / 2)
        self.limit = (param_mid - param_extent, param_mid + param_extent)

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        factor = 1.0 + random.uniform(self.limit[0], self.limit[1])
        return [Brightness._apply_brightness(elem, factor) for elem in data]

    @staticmethod
    def _apply_brightness(data: np.ndarray, factor: float) -> np.ndarray:
        im = Image.fromarray(data)
        im = ImageEnhance.Brightness(im).enhance(factor)
        return np.array(im)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
import random
from typing import Any, Dict, List, Union

import numpy as np

from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable


@traceable()
class OneOf(NumpyOp):
    """Perform one of several possible NumpyOps.

    Args:
        *numpy_ops: A list of ops to choose between with uniform probability.
    """
    def __init__(self, *numpy_ops: NumpyOp) -> None:
        inputs = numpy_ops[0].inputs
        outputs = numpy_ops[0].outputs
        mode = numpy_ops[0].mode
        ds_id = numpy_ops[0].ds_id
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.in_list = numpy_ops[0].in_list
        self.out_list = numpy_ops[0].out_list
        for op in numpy_ops[1:]:
            assert inputs == op.inputs, "All ops within a OneOf must share the same inputs"
            assert self.in_list == op.in_list, "All ops within OneOf must share the same input configuration"
            assert outputs == op.outputs, "All ops within a OneOf must share the same outputs"
            assert self.out_list == op.out_list, "All ops within OneOf must share the same output configuration"
            assert mode == op.mode, "All ops within a OneOf must share the same mode"
            assert ds_id == op.ds_id, "All ops within a OneOf must share the same ds_id"
        self.ops = numpy_ops

    def __getstate__(self) -> Dict[str, List[Dict[Any, Any]]]:
        return {'ops': [elem.__getstate__() if hasattr(elem, '__getstate__') else {} for elem in self.ops]}

    def set_rua_level(self, magnitude_coef: float) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).

        Raises:
            AttributeError: If ops don't have a 'set_rua_level' method.
        """
        for op in self.ops:
            if hasattr(op, "set_rua_level") and inspect.ismethod(getattr(op, "set_rua_level")):
                op.set_rua_level(magnitude_coef=magnitude_coef)
            else:
                raise AttributeError(
                    "RUA Augmentations should have a 'set_rua_level' method but it's not present in Op: {}".format(
                        op.__class__.__name__))

    def forward(self, data: Union[np.ndarray, List[np.ndarray]],
                state: Dict[str, Any]) -> Union[np.ndarray, List[np.ndarray]]:
        """Execute a randomly selected op from the list of `numpy_ops`.

        Args:
            data: The information to be passed to one of the wrapped operators.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The `data` after application of one of the available numpyOps.
        """
        return random.choice(self.ops).forward(data, state)



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import argparse
import sys
from typing import Any, Dict, List

from fastestimator.summary.history import HistoryReader, delete, update_settings


def history_basic(args: Dict[str, Any], unknown: List[str]) -> None:
    """A method to query FE history using CLI-provided arguments.

    Args:
        args: The arguments to be fed to the read_sql() method.
        unknown: Any cli arguments not matching known inputs for the read_sql() method.

    Raises:
        SystemExit: If `unknown` arguments were provided by the user.
    """
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    with HistoryReader() as reader:
        reader.read_basic(limit=args['limit'],
                          interactive=args['interactive'],
                          include_args=args['include_args'],
                          errors=args['errors'],
                          include_pk=args['include_pk'],
                          include_traces=args['include_traces'],
                          include_features=args['include_features'],
                          include_datasets=args['include_datasets'],
                          include_pipeline=args['include_pipeline'],
                          include_network=args['include_network'],
                          as_csv=args['csv'])


def history_sql(args: Dict[str, Any], unknown: List[str]) -> None:
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    with HistoryReader() as reader:
        reader.read_sql(query=args['query'], as_csv=args['csv'], interactive=args['interactive'])


def clear_history(args: Dict[str, Any], unknown: List[str]) -> None:
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    delete(n_keep=args['retain'])


def settings(args: Dict[str, Any], unknown: List[str]) -> None:
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    update_settings(n_keep=args['keep'], n_keep_logs=args['keep_logs'])


def configure_history_parser(subparsers: argparse._SubParsersAction) -> None:
    """Add a history parser to an existing argparser.

    Args:
        subparsers: The parser object to be appended to.
    """
    parser = subparsers.add_parser('history',
                                   description='View prior FE training histories',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                   allow_abbrev=False)
    parser.add_argument('--limit', metavar='L', type=int, help="How many entries to return", default=15)
    parser.add_argument('--interactive',
                        dest='interactive',
                        help="Whether to run an interactive session which lets you look up detailed information",
                        action='store_true',
                        default=False)
    parser.add_argument('--args',
                        dest='include_args',
                        help="Whether to return a list of the args used to invoke the training",
                        action='store_true',
                        default=False)
    parser.add_argument('--errors',
                        dest='errors',
                        help="Whether to focus on failed trainings and include extra error information",
                        action='store_true',
                        default=False)
    parser.add_argument('--features',
                        dest='include_features',
                        help="Whether to return a list of the FE features used during each training",
                        action='store_true',
                        default=False)
    parser.add_argument('--datasets',
                        dest='include_datasets',
                        help="Whether to return a list of the datasets used during each training",
                        action='store_true',
                        default=False)
    parser.add_argument('--pipeline',
                        dest='include_pipeline',
                        help="Whether to return a list of the pipeline ops used during each training",
                        action='store_true',
                        default=False)
    parser.add_argument('--network',
                        dest='include_network',
                        help="Whether to return a list of the network ops used during each training",
                        action='store_true',
                        default=False)
    parser.add_argument('--traces',
                        dest='include_traces',
                        help="Whether to return a list of the traces used during each training",
                        action='store_true',
                        default=False)
    parser.add_argument('--pks',
                        dest='include_pk',
                        help="Whether to return the database primary keys of each entry",
                        action='store_true',
                        default=False)
    parser.add_argument('--csv',
                        dest='csv',
                        help='Print the response as a csv rather than a formatted table',
                        action='store_true',
                        default=False)
    sp = parser.add_subparsers()
    sql_parser = sp.add_parser('sql',
                               description='Perform a raw SQL query against the history database',
                               formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                               allow_abbrev=False)
    sql_parser.add_argument('query',
                            metavar='<Query>',
                            type=str,
                            help="ex: fastestimator history sql 'SELECT * FROM history'")
    sql_parser.add_argument('--interactive',
                            dest='interactive',
                            help="Whether to run an interactive session which lets you look up detailed information",
                            action='store_true',
                            default=False)
    sql_parser.add_argument('--csv',
                            dest='csv',
                            help='Print the response as a csv rather than a formatted table',
                            action='store_true',
                            default=False)
    sql_parser.set_defaults(func=history_sql)
    clear_parser = sp.add_parser('clear',
                                 description='Clear out old history entries to save space',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                 allow_abbrev=False)
    clear_parser.add_argument('retain',
                              metavar='N',
                              nargs='?',
                              type=int,
                              help="How many of the most recent entries to keep",
                              default=20)
    clear_parser.set_defaults(func=clear_history)
    settings_parser = sp.add_parser('settings',
                                    description="Modify history settings, such as how many logs to retain",
                                    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                    allow_abbrev=False)
    settings_parser.add_argument('--keep',
                                 metavar='K',
                                 type=int,
                                 help="How many of the most recent history entries to keep")
    settings_parser.add_argument('--keep_logs',
                                 metavar='L',
                                 type=int,
                                 help="How many of the most recent log entries to keep")
    settings_parser.set_defaults(func=settings)
    parser.set_defaults(func=history_basic)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Tuple, Union

from albumentations.augmentations.transforms import ISONoise as ISONoiseAlb

from fastestimator.op.numpyop.univariate.univariate import ImageOnlyAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class ISONoise(ImageOnlyAlbumentation):
    """Apply camera sensor noise.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        color_shift: Variance range for color hue change. Measured as a fraction of 360 degree Hue angle in the HLS
            colorspace.
        intensity: Multiplicative factor that controls the strength of color and luminace noise.

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 color_shift: Tuple[float, float] = (0.01, 0.05),
                 intensity: Tuple[float, float] = (0.1, 0.5)):
        super().__init__(ISONoiseAlb(color_shift=color_shift, intensity=intensity, always_apply=True),
                         inputs=inputs,
                         outputs=outputs,
                         mode=mode,
                         ds_id=ds_id)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, List, Optional, Set, TypeVar, Union

import tensorflow as tf
import tensorflow_probability as tfp
import torch

from fastestimator.backend.cast import cast
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class OneOf(TensorOp):
    """Perform one of several possible TensorOps.

    Args:
        *tensor_ops: A list of ops to choose between with uniform probability.
    """
    def __init__(self, *tensor_ops: TensorOp) -> None:
        inputs = tensor_ops[0].inputs
        outputs = tensor_ops[0].outputs
        mode = tensor_ops[0].mode
        ds_id = tensor_ops[0].ds_id
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.in_list = tensor_ops[0].in_list
        self.out_list = tensor_ops[0].out_list
        for op in tensor_ops[1:]:
            assert inputs == op.inputs, "All ops within a OneOf must share the same inputs"
            assert self.in_list == op.in_list, "All ops within OneOf must share the same input configuration"
            assert outputs == op.outputs, "All ops within a OneOf must share the same outputs"
            assert self.out_list == op.out_list, "All ops within OneOf must share the same output configuration"
            assert mode == op.mode, "All ops within a OneOf must share the same mode"
            assert ds_id == op.ds_id, "All ops within a OneOf must share the same ds_id"
        self.ops = tensor_ops
        self.prob_fn = None
        self.invoke_fn = None

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        for op in self.ops:
            op.build(framework, device)
        if framework == 'tf':
            self.prob_fn = tfp.distributions.Uniform(low=0, high=len(self.ops))
            self.invoke_fn = lambda idx, data, state: tf.switch_case(idx, [lambda: op.forward(data, state) for op in
                                                                           self.ops])
        elif framework == 'torch':
            self.prob_fn = torch.distributions.uniform.Uniform(low=0, high=len(self.ops))
            self.invoke_fn = lambda idx, data, state: self.ops[idx].forward(data, state)
        else:
            raise ValueError("unrecognized framework: {}".format(framework))

    def get_fe_loss_keys(self) -> Set[str]:
        return set.union(*[op.get_fe_loss_keys() for op in self.ops])

    def get_fe_models(self) -> Set[Model]:
        return set.union(*[op.get_fe_models() for op in self.ops])

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        resp = None
        for op in self.ops:
            resp = resp or op.fe_retain_graph(retain)
        return resp

    def __getstate__(self) -> Dict[str, List[Dict[Any, Any]]]:
        return {'ops': [elem.__getstate__() if hasattr(elem, '__getstate__') else {} for elem in self.ops]}

    def forward(self, data: Union[Tensor, List[Tensor]], state: Dict[str, Any]) -> Union[Tensor, List[Tensor]]:
        """Execute a randomly selected op from the list of `numpy_ops`.

        Args:
            data: The information to be passed to one of the wrapped operators.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The `data` after application of one of the available numpyOps.
        """
        idx = cast(self.prob_fn.sample(), dtype='int32')
        return self.invoke_fn(idx, data, state)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Optional, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.clip_by_value import clip_by_value
from fastestimator.backend.get_gradient import get_gradient
from fastestimator.backend.reduce_max import reduce_max
from fastestimator.backend.reduce_min import reduce_min
from fastestimator.backend.sign import sign
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


@traceable()
class FGSM(TensorOp):
    """Create an adversarial sample from input data using the Fast Gradient Sign Method.

    See https://arxiv.org/abs/1412.6572 for an explanation of adversarial attacks.

    Args:
        data: Key of the input to be attacked.
        loss: Key of the loss value to use for gradient computation.
        outputs: The key under which to save the output.
        epsilon: The strength of the perturbation to use in the attack.
        clip_low: a minimum value to clip the output by (defaults to min value of data when set to None).
        clip_high: a maximum value to clip the output by (defaults to max value of data when set to None).
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 data: str,
                 loss: str,
                 outputs: str,
                 epsilon: float = 0.01,
                 clip_low: Optional[float] = None,
                 clip_high: Optional[float] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=[data, loss], outputs=outputs, mode=mode, ds_id=ds_id)
        self.epsilon = epsilon
        self.clip_low = clip_low
        self.clip_high = clip_high
        self.retain_graph = True

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        if retain is not None:
            self.retain_graph = retain
        return self.retain_graph

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> Tensor:
        data, loss = data
        grad = get_gradient(target=loss, sources=data, tape=state['tape'], retain_graph=self.retain_graph)
        adverse_data = clip_by_value(data + self.epsilon * sign(grad),
                                     min_value=self.clip_low or reduce_min(data),
                                     max_value=self.clip_high or reduce_max(data))
        return adverse_data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestPow(unittest.TestCase):
    def test_pow_np_input(self):
        n = np.array([-2, 7, -19])
        obj1 = fe.backend.pow(n, 2)
        obj2 = np.array([4, 49, 361])
        self.assertTrue(is_equal(obj1, obj2))

    def test_pow_tf_input(self):
        t = tf.constant([-2, 7, -19])
        obj1 = fe.backend.pow(t, 2)
        obj2 = tf.constant([4, 49, 361])
        self.assertTrue(is_equal(obj1, obj2))

    def test_pow_torch_input(self):
        t = torch.tensor([-2, 7, -19])
        obj1 = fe.backend.pow(t, 2)
        obj2 = torch.tensor([4, 49, 361])
        self.assertTrue(is_equal(obj1, obj2))



import datetime
import os
import re
import sys

from setuptools import find_packages, setup

is_nightly = os.environ.get('FASTESTIMATOR_IS_NIGHTLY', None)
if is_nightly is not None:
    sys.stderr.write("Using '%s=%s' environment variable!\n" % ('FASTESTIMATOR_IS_NIGHTLY', is_nightly))


def get_version():
    path = os.path.dirname(__file__)
    version_re = re.compile(r'''__version__ = ['"](.+)['"]''')
    with open(os.path.join(path, 'fastestimator', '__init__.py')) as f:
        init = f.read()

    now = datetime.datetime.now()
    version = version_re.search(init).group(1)
    if is_nightly:
        return "{}.dev{:04}{:02}{:02}{:02}{:02}".format(version, now.year, now.month, now.day, now.hour, now.minute)
    else:
        return version


def get_name():
    if is_nightly:
        return "fastestimator-nightly"
    else:
        return "fastestimator"


def get_dependency():
    dependencies = [
        'numpy',
        'albumentations==0.5.2',
        'pyfiglet',
        'opencv-python',
        'pandas',
        'sklearn',
        'wget',
        'pillow',
        'seaborn',
        'matplotlib==3.3.3',
        'requests',
        'tqdm',
        'h5py==2.10.0',
        'jsonpickle',
        'python-docx',
        'scipy==1.5.4',
        'PyLaTeX==1.4.1',
        'natsort==7.1.0',
        'tensorflow_probability==0.12.1',
        'tensorflow-addons==0.12.1',
        'transformers==4.6.0',
        'pytorch-model-summary==0.1.2',
        'graphviz==0.15',
        'hiddenlayer==0.3',
        'pydot==1.4.1',
        'dot2tex==2.11.3',
        'gdown==3.12.0',
        'PySocks==1.7.1',
        'uncertainty-calibration==0.0.8',
        'dill==0.3.3',
        'scikit-image==0.17.2',
        'prettytable==2.2.1'
    ]
    if os.name == "nt":
        dependencies.append(
            "pycocotools @ git+https://github.com/philferriere/cocoapi.git#egg=pycocotools&subdirectory=PythonAPI")
    else:
        dependencies.append('pycocotools-fix')
    return dependencies


setup(
    entry_points={"console_scripts": ["fastestimator = fastestimator.cli.main:main"]},
    name=get_name(),
    version=get_version(),
    description="Deep learning framework",
    packages=find_packages(),
    package_dir={'': '.'},
    long_description="FastEstimator is a high-level deep learning API. With the help of FastEstimator, you can easily \
                    build a high-performance deep learning model and run it anywhere.",
    author="FastEstimator Dev",
    url='https://github.com/fastestimator/fastestimator',
    license="Apache License 2.0",
    keywords="fastestimator tensorflow pytorch",
    classifiers=["License :: OSI Approved :: Apache Software License", "Programming Language :: Python :: 3"],
    # Declare minimal set for installation
    install_requires=get_dependency(),
    # Declare extra set for installation
    extras_require={})



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence, TypeVar, Union

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend.abs import abs
from fastestimator.backend.argmax import argmax
from fastestimator.backend.clip_by_value import clip_by_value
from fastestimator.backend.percentile import percentile
from fastestimator.backend.random_normal_like import random_normal_like
from fastestimator.backend.random_uniform_like import random_uniform_like
from fastestimator.backend.reduce_max import reduce_max
from fastestimator.backend.reduce_min import reduce_min
from fastestimator.backend.reduce_sum import reduce_sum
from fastestimator.network import Network
from fastestimator.op.tensorop.gather import Gather
from fastestimator.op.tensorop.gradient.gradient import GradientOp
from fastestimator.op.tensorop.gradient.watch import Watch
from fastestimator.op.tensorop.model.model import ModelOp
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list, to_number

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class SaliencyNet:
    """A class to generate saliency masks from a given model.

    Args:
        model: The model, compiled with fe.build, which is to be inspected.
        model_inputs: The key(s) corresponding to the model inputs within the data dictionary.
        model_outputs: The key(s) corresponding to the model outputs which are written into the data dictionary.
        outputs: The keys(s) under which to write the generated saliency images.
    """
    def __init__(self,
                 model: Model,
                 model_inputs: Union[str, Sequence[str]],
                 model_outputs: Union[str, Sequence[str]],
                 outputs: Union[str, List[str]] = "saliency"):
        mode = "test"
        self.model_op = ModelOp(model=model, mode=mode, inputs=model_inputs, outputs=model_outputs, trainable=False)
        self.outputs = to_list(outputs)
        self.mode = mode
        self.gather_keys = ["SaliencyNet_Target_Index_{}".format(key) for key in self.model_outputs]
        self.network = Network(ops=[
            Watch(inputs=self.model_inputs, mode=mode),
            self.model_op,
            Gather(inputs=self.model_outputs,
                   indices=self.gather_keys,
                   outputs=["SaliencyNet_Intermediate_{}".format(key) for key in self.model_outputs],
                   mode=mode),
            GradientOp(inputs=self.model_inputs,
                       finals=["SaliencyNet_Intermediate_{}".format(key) for key in self.model_outputs],
                       outputs=deepcopy(self.outputs),
                       mode=mode),
        ])

    @property
    def model_inputs(self):
        return deepcopy(self.model_op.inputs)

    @property
    def model_outputs(self):
        return deepcopy(self.model_op.outputs)

    @staticmethod
    def _convert_for_visualization(tensor: Tensor, tile: int = 99) -> np.ndarray:
        """Modify the range of data in a given input `tensor` to be appropriate for visualization.

        Args:
            tensor: Input masks, whose channel values are to be reduced by absolute value summation.
            tile: The percentile [0-100] used to set the max value of the image.

        Returns:
            A (batch X width X height) image after visualization clipping is applied.
        """
        if isinstance(tensor, torch.Tensor):
            channel_axis = 1
        else:
            channel_axis = -1
        flattened_mask = reduce_sum(abs(tensor), axis=channel_axis, keepdims=True)

        non_batch_axes = list(range(len(flattened_mask.shape)))[1:]

        vmax = percentile(flattened_mask, tile, axis=non_batch_axes, keepdims=True)
        vmin = reduce_min(flattened_mask, axis=non_batch_axes, keepdims=True)

        return clip_by_value((flattened_mask - vmin) / (vmax - vmin), 0, 1)

    def get_masks(self, batch: Dict[str, Any]) -> Dict[str, Union[Tensor, np.ndarray]]:
        """Generates greyscale saliency mask(s) from a given `batch` of data.

        Args:
            batch: A batch of input data to be fed to the model.

        Returns:
            The model's classification decisions and greyscale saliency mask(s) for the given `batch` of data.
        """
        # Shallow copy batch since we're going to modify its contents later
        batch = {key: val for key, val in batch.items()}
        grads_and_preds = self._get_mask(batch)
        for key in self.outputs:
            grads_and_preds[key] = self._convert_for_visualization(grads_and_preds[key])
        return grads_and_preds

    def _get_mask(self, batch: Dict[str, Any]) -> Dict[str, Tensor]:
        """Generates raw saliency mask(s) from a given `batch` of data.

        This method assumes that the Network is already loaded.

        Args:
            batch: A batch of input data to be fed to the model.

        Returns:
            The model outputs and the raw saliency mask(s) for the given `batch` of data. Model predictions are reduced
            via argmax.
        """
        for key in self.gather_keys:
            # If there's no target key, use an empty array which will cause the max-likelihood class to be selected
            batch.setdefault(key, [])
        prediction = self.network.transform(data=batch, mode=self.mode)
        for key in self.model_outputs:
            prediction[key] = argmax(prediction[key], axis=1)
        return prediction

    def _get_integrated_masks(self, batch: Dict[str, Any], nsamples: int = 25) -> Dict[str, Tensor]:
        """Generates raw integrated saliency mask(s) from a given `batch` of data.

        This method assumes that the Network is already loaded.

        Args:
            batch: A batch of input data to be fed to the model.
            nsamples: How many samples to consider during integration.

        Returns:
            The raw integrated saliency mask(s) for the given `batch` of data.
        """
        model_inputs = [batch[ins] for ins in self.model_inputs]

        # Use a random uniform baseline as advised in https://distill.pub/2020/attribution-baselines/
        input_baselines = [
            random_uniform_like(ins, minval=reduce_min(ins), maxval=reduce_max(ins)) for ins in model_inputs
        ]
        input_diffs = [
            model_input - input_baseline for model_input, input_baseline in zip(model_inputs, input_baselines)
        ]

        response = {}

        for alpha in np.linspace(0.0, 1.0, nsamples):
            noisy_batch = {key: batch[key] for key in self.gather_keys}
            for idx, input_name in enumerate(self.model_inputs):
                x_step = input_baselines[idx] + alpha * input_diffs[idx]
                noisy_batch[input_name] = x_step
            grads_and_preds = self._get_mask(noisy_batch)
            for key in self.outputs:
                if key in response:
                    response[key] += grads_and_preds[key]
                else:
                    response[key] = grads_and_preds[key]

        for key in self.outputs:
            grad = response[key]
            for diff in input_diffs:
                grad = grad * diff
            response[key] = grad

        return response

    def get_smoothed_masks(self,
                           batch: Dict[str, Any],
                           stdev_spread: float = .15,
                           nsamples: int = 25,
                           nintegration: Optional[int] = None,
                           magnitude: bool = True) -> Dict[str, Union[Tensor, np.ndarray]]:
        """Generates smoothed greyscale saliency mask(s) from a given `batch` of data.

        Args:
            batch: An input batch of data.
            stdev_spread: Amount of noise to add to the input, as fraction of the total spread (x_max - x_min).
            nsamples: Number of samples to average across to get the smooth gradient.
            nintegration: Number of samples to compute when integrating (None to disable).
            magnitude: If true, computes the sum of squares of gradients instead of just the sum.

        Returns:
            Greyscale saliency mask(s) smoothed via the SmoothGrad method.
        """
        # Shallow copy batch since we're going to modify its contents later
        batch = {key: val for key, val in batch.items()}
        model_inputs = [batch[ins] for ins in self.model_inputs]
        stdevs = [to_number(stdev_spread * (reduce_max(ins) - reduce_min(ins))).item() for ins in model_inputs]

        # Adding noise to the image might cause the max likelihood class value to change, so need to keep track of
        # which class we're comparing to
        response = self._get_mask(batch)
        for gather_key, output_key in zip(self.gather_keys, self.model_outputs):
            batch[gather_key] = response[output_key]

        if magnitude:
            for key in self.outputs:
                response[key] = response[key] * response[key]

        for _ in range(nsamples - 1):
            noisy_batch = {key: batch[key] for key in self.gather_keys}
            clean_batch = {key: val for key, val in noisy_batch.items()}
            for idx, input_name in enumerate(self.model_inputs):
                noise = random_normal_like(model_inputs[idx], std=stdevs[idx])
                x_plus_noise = model_inputs[idx] + noise
                clean_batch[input_name] = model_inputs[idx]
                noisy_batch[input_name] = x_plus_noise
            grads_and_preds = self._get_mask(noisy_batch) if not nintegration else self._get_integrated_masks(
                clean_batch, nsamples=nintegration)  # Integration introduces its own noise pattern
            for name in self.outputs:
                grad = grads_and_preds[name]
                if magnitude:
                    response[name] += grad * grad
                else:
                    response[name] += grad
        for key in self.outputs:
            grad = response[key]
            response[key] = self._convert_for_visualization(grad / nsamples)
        return response

    def get_integrated_masks(self, batch: Dict[str, Any], nsamples: int = 25) -> Dict[str, Union[Tensor, np.ndarray]]:
        """Generates integrated greyscale saliency mask(s) from a given `batch` of data.

        See https://arxiv.org/abs/1703.01365 for background on the IntegratedGradient method.

        Args:
            batch: An input batch of data.
            nsamples: Number of samples to average across to get the integrated gradient.

        Returns:
            Greyscale saliency masks smoothed via the IntegratedGradient method.
        """
        # Shallow copy batch since we're going to modify its contents later
        batch = {key: val for key, val in batch.items()}

        # Performing integration might cause the max likelihood class value to change, so need to keep track of
        # which class we're comparing to
        response = self._get_mask(batch)
        for gather_key, output_key in zip(self.gather_keys, self.model_outputs):
            batch[gather_key] = response[output_key]

        response.update(self._get_integrated_masks(batch, nsamples=nsamples))
        for key in self.outputs:
            response[key] = self._convert_for_visualization(response[key])

        return response



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.op.numpyop.univariate import Tokenize
from fastestimator.test.unittest_util import is_equal


class TestTokenize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.single_input = ['This is a function to test tokenize op']
        cls.single_output = [['This', 'is', 'a', 'function', 'to', 'test', 'tokenize', 'op']]
        cls.multi_input = ['Now define a new list', 'that contains multiple sentences']
        cls.multi_output = [['Now', 'define', 'a', 'new', 'list'], ['that', 'contains', 'multiple', 'sentences']]
        cls.lower_case_input = ['To Test Lowercase parameter']
        cls.tokenize_fn_output = [['THIS', 'IS', 'A', 'FUNCTION', 'TO', 'TEST', 'TOKENIZE', 'OP']]

    def tokenize_fn(self, seq):
        seq = seq.upper().split()
        return seq

    def test_single_input(self):
        op = Tokenize(inputs='x', outputs='x')
        data = op.forward(data=self.single_input, state={})
        self.assertTrue(is_equal(data, self.single_output))

    def test_single_input_tokenize_function(self):
        op = Tokenize(inputs='x', outputs='x', tokenize_fn=self.tokenize_fn)
        data = op.forward(data=self.single_input, state={})
        self.assertTrue(is_equal(data, self.tokenize_fn_output))

    def test_multi_input(self):
        op = Tokenize(inputs='x', outputs='x')
        data = op.forward(data=self.multi_input, state={})
        self.assertTrue(is_equal(data, self.multi_output))

    def test_lower_case(self):
        op = Tokenize(inputs='x', outputs='x', to_lower_case=True)
        data = op.forward(data=self.lower_case_input, state={})
        self.assertTrue(is_equal(data, [['to', 'test', 'lowercase', 'parameter']]))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def reshape(tensor: Tensor, shape: List[int]) -> Tensor:
    """Reshape a `tensor` to conform to a given shape.

    This method can be used with Numpy data:
    ```python
    n = np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
    b = fe.backend.reshape(n, shape=[-1])  # [1, 2, 3, 4, 5, 6, 7, 8]
    b = fe.backend.reshape(n, shape=[2, 4])  # [[1, 2, 3, 4], [5, 6, 7, 8]]
    b = fe.backend.reshape(n, shape=[4, 2])  # [[1, 2], [3, 4], [5, 6], [7, 8]]
    b = fe.backend.reshape(n, shape=[2, 2, 2, 1])  # [[[[1], [2]], [[3], [4]]], [[[5], [6]], [[7], [8]]]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
    b = fe.backend.reshape(t, shape=[-1])  # [1, 2, 3, 4, 5, 6, 7, 8]
    b = fe.backend.reshape(t, shape=[2, 4])  # [[1, 2, 3, 4], [5, 6, 7, 8]]
    b = fe.backend.reshape(t, shape=[4, 2])  # [[1, 2], [3, 4], [5, 6], [7, 8]]
    b = fe.backend.reshape(t, shape=[2, 2, 2, 1])  # [[[[1], [2]], [[3], [4]]], [[[5], [6]], [[7], [8]]]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
    b = fe.backend.reshape(p, shape=[-1])  # [1, 2, 3, 4, 5, 6, 7, 8]
    b = fe.backend.reshape(p, shape=[2, 4])  # [[1, 2, 3, 4], [5, 6, 7, 8]]
    b = fe.backend.reshape(p, shape=[4, 2])  # [[1, 2], [3, 4], [5, 6], [7, 8]]
    b = fe.backend.reshape(p, shape=[2, 2, 2, 1])  # [[[[1], [2]], [[3], [4]]], [[[5], [6]], [[7], [8]]]]
    ```

    Args:
        tensor: The input value.
        shape: The new shape of the tensor. At most one value may be -1 which indicates that whatever values are left
            should be packed into that axis.

    Returns:
        The reshaped `tensor`.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        return tf.reshape(tensor, shape=shape)
    elif isinstance(tensor, torch.Tensor):
        return torch.reshape(tensor, shape=shape)
    elif isinstance(tensor, np.ndarray):
        return np.reshape(tensor, shape)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))
